from typing import Any, Optional, Type, Union, Dict, TYPE_CHECKING, Sequence, Tuple

from dlt.common.destination import Destination, DestinationCapabilitiesContext
from dlt.common.arithmetics import DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE
from dlt.common.data_writers.escape import escape_dremio_identifier

from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.exceptions import TerminalValueError
from dlt.common.schema.typing import TColumnSchema, TColumnType
from dlt.common.typing import TLoaderFileFormat
from dlt.destinations.type_mapping import TypeMapperImpl
from dlt.destinations.impl.dremio.configuration import (
    DremioCredentials,
    DremioClientConfiguration,
)

if TYPE_CHECKING:
    from dlt.destinations.impl.dremio.dremio import DremioClient


class DremioTypeMapper(TypeMapperImpl):
    BIGINT_PRECISION = 19
    sct_to_unbound_dbt = {
        "json": "VARCHAR",
        "text": "VARCHAR",
        "double": "DOUBLE",
        "bool": "BOOLEAN",
        "date": "DATE",
        "timestamp": "TIMESTAMP",
        "bigint": "BIGINT",
        "binary": "VARBINARY",
        "time": "TIME",
    }

    sct_to_dbt = {
        "decimal": "DECIMAL(%i,%i)",
        "wei": "DECIMAL(%i,%i)",
    }

    dbt_to_sct = {
        "VARCHAR": "text",
        "DOUBLE": "double",
        "FLOAT": "double",
        "BOOLEAN": "bool",
        "DATE": "date",
        "TIMESTAMP": "timestamp",
        "VARBINARY": "binary",
        "BINARY": "binary",
        "BINARY VARYING": "binary",
        "VARIANT": "json",
        "TIME": "time",
        "BIGINT": "bigint",
        "DECIMAL": "decimal",
    }

    def ensure_supported_type(
        self,
        column: TColumnSchema,
        table: PreparedTableSchema,
        loader_file_format: TLoaderFileFormat,
    ) -> None:
        if loader_file_format == "insert_values":
            return
        if loader_file_format == "parquet":
            # binary not supported on parquet if precision is set
            if column.get("precision") is not None and column["data_type"] == "binary":
                raise TerminalValueError(
                    "Dremio cannot load fixed width `binary` columns from parquet files. Switch to"
                    " other file format or use binary columns without precision.",
                    "binary",
                    column["data_type"],
                )

    def from_destination_type(
        self, db_type: str, precision: Optional[int] = None, scale: Optional[int] = None
    ) -> TColumnType:
        if db_type == "DECIMAL":
            if (precision, scale) == self.capabilities.wei_precision:
                return dict(data_type="wei")
            return dict(data_type="decimal", precision=precision, scale=scale)
        return super().from_destination_type(db_type, precision, scale)


class dremio(Destination[DremioClientConfiguration, "DremioClient"]):
    spec = DremioClientConfiguration

    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        caps = DestinationCapabilitiesContext()
        caps.preferred_loader_file_format = None
        caps.supported_loader_file_formats = []
        caps.preferred_staging_file_format = "parquet"
        caps.supported_staging_file_formats = ["jsonl", "parquet", "model"]
        caps.escape_identifier = escape_dremio_identifier
        caps.type_mapper = DremioTypeMapper
        # all identifiers are case insensitive but are stored as is
        # https://docs.dremio.com/current/sonar/data-sources
        caps.has_case_sensitive_identifiers = False
        caps.decimal_precision = (DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE)
        caps.wei_precision = (DEFAULT_NUMERIC_PRECISION, 0)
        caps.max_identifier_length = 255
        caps.max_column_identifier_length = 255
        caps.max_query_length = 2 * 1024 * 1024
        caps.is_max_query_length_in_bytes = True
        caps.max_text_data_type_length = 16 * 1024 * 1024
        caps.is_max_text_data_type_length_in_bytes = True
        caps.supports_transactions = False
        caps.supports_ddl_transactions = False
        caps.alter_add_multi_column = True
        caps.supports_clone_table = False
        caps.supports_multiple_statements = False
        caps.timestamp_precision = 3
        caps.max_timestamp_precision = 3
        caps.supported_merge_strategies = ["delete-insert", "scd2"]
        caps.supported_replace_strategies = ["truncate-and-insert", "insert-from-staging"]
        caps.enforces_nulls_on_alter = False
        caps.sqlglot_dialect = "presto"
        caps.supports_tz_aware_datetime = False

        return caps

    @property
    def client_class(self) -> Type["DremioClient"]:
        from dlt.destinations.impl.dremio.dremio import DremioClient

        return DremioClient

    def __init__(
        self,
        credentials: Union[DremioCredentials, Dict[str, Any], str] = None,
        staging_data_source: str = None,
        destination_name: str = None,
        environment: str = None,
        **kwargs: Any,
    ) -> None:
        """Configure the Dremio destination to use in a pipeline.

        All arguments provided here supersede other configuration sources such as environment variables and dlt config files.

        Args:
            credentials (Union[DremioCredentials, Dict[str, Any], str], optional): Credentials to connect to the dremio database. Can be an instance of `DremioCredentials` or
                a connection string in the format `dremio://user:password@host:port/database`
            staging_data_source (str, optional): The name of the "Object Storage" data source in Dremio containing the s3 bucket
            destination_name (str, optional): Name of the destination, can be used in config section to differentiate between multiple of the same type
            environment (str, optional): Environment of the destination
            **kwargs (Any): Additional arguments passed to the destination config
        """
        super().__init__(
            credentials=credentials,
            staging_data_source=staging_data_source,
            destination_name=destination_name,
            environment=environment,
            **kwargs,
        )


dremio.register()



from typing import List

from dlt.common import logger
from dlt.common.metrics import DataWriterMetrics
from dlt.common.schema import TSchemaUpdate
from dlt.common.storages.load_package import ParsedLoadJobFileName

from dlt.normalize.items_normalizers.base import ItemsNormalizer


class FileImportNormalizer(ItemsNormalizer):
    def __call__(self, extracted_items_file: str, root_table_name: str) -> List[TSchemaUpdate]:
        self._maybe_cancel()
        logger.info(
            f"Table {root_table_name} {self.item_storage.writer_spec.file_format} file"
            f" {extracted_items_file} will be directly imported without normalization"
        )
        completed_columns = self.schema.get_table_columns(root_table_name)
        if not completed_columns:
            logger.warning(
                f"Table {root_table_name} has no completed columns for imported file"
                f" {extracted_items_file} and will not be created! Pass column hints to the"
                " resource or with dlt.mark.with_hints or create the destination table yourself."
            )
        with self.normalize_storage.extracted_packages.storage.open_file(
            extracted_items_file, "rb"
        ) as f:
            # TODO: sniff the schema depending on a file type
            file_metrics = DataWriterMetrics(extracted_items_file, 0, f.tell(), 0, 0)
        parts = ParsedLoadJobFileName.parse(extracted_items_file)
        self.item_storage.import_items_file(
            self.load_id,
            self.schema.name,
            parts.table_name,
            self.normalize_storage.extracted_packages.storage.make_full_path(extracted_items_file),
            file_metrics,
        )
        return []



from typing import Sequence

from dlt.common.destination.exceptions import (
    DestinationTerminalException,
    DestinationTransientException,
    DestinationUndefinedEntity,
    DestinationException,
)
from dlt.common.destination.client import TLoadJobState


class DatabaseException(DestinationException):
    def __init__(self, dbapi_exception: Exception) -> None:
        self.dbapi_exception = dbapi_exception
        super().__init__(dbapi_exception)


class DatabaseTerminalException(DestinationTerminalException, DatabaseException):
    def __init__(self, dbapi_exception: Exception) -> None:
        super().__init__(dbapi_exception)


class DatabaseUndefinedRelation(DestinationUndefinedEntity, DatabaseException):
    def __init__(self, dbapi_exception: Exception) -> None:
        super().__init__(dbapi_exception)


class DatabaseTransientException(DestinationTransientException, DatabaseException):
    def __init__(self, dbapi_exception: Exception) -> None:
        super().__init__(dbapi_exception)


class DestinationConnectionError(DestinationTransientException):
    def __init__(
        self, client_type: str, dataset_name: str, reason: str, inner_exc: Exception
    ) -> None:
        self.client_type = client_type
        self.dataset_name = dataset_name
        self.inner_exc = inner_exc
        super().__init__(
            f"Connection with `{client_type=:}` to `{dataset_name=:}` failed. Please check if"
            " you configured the credentials at all and provided the right credentials values. You"
            " can be also denied access or your internet connection may be down. The actual reason"
            f" given is: {reason}"
        )


class LoadClientNotConnected(DestinationTransientException):
    def __init__(self, client_type: str, dataset_name: str) -> None:
        self.client_type = client_type
        self.dataset_name = dataset_name
        super().__init__(
            f"Connection with `{client_type=:}` to `{dataset_name=:}` is closed. Open the"
            " connection with `client.open_connection` or with the `with client:` context manager"
        )


class DestinationSchemaWillNotUpdate(DestinationTerminalException):
    def __init__(self, table_name: str, columns: Sequence[str], msg: str) -> None:
        self.table_name = table_name
        self.columns = columns
        super().__init__(
            f"Schema for table `{table_name}` column(s) `{columns}` will not update: {msg}"
        )


class LoadJobNotExistsException(DestinationTerminalException):
    def __init__(self, job_id: str) -> None:
        super().__init__(f"Job with {job_id=:} not found")


class LoadJobTerminalException(DestinationTerminalException):
    def __init__(self, file_path: str, message: str) -> None:
        super().__init__(f"Job with `{file_path=:}` encountered unrecoverable problem: {message}")


class LoadJobInvalidStateTransitionException(DestinationTerminalException):
    def __init__(self, from_state: TLoadJobState, to_state: TLoadJobState) -> None:
        self.from_state = from_state
        self.to_state = to_state
        super().__init__(f"Load job cannot transition from `{from_state}` to `{to_state}`")


class LoadJobFileTooBig(DestinationTerminalException):
    def __init__(self, file_name: str, max_size: int) -> None:
        super().__init__(
            f"File `{file_name}` exceeds `{max_size=:}` and cannot be loaded. Split the file and"
            " try again."
        )


class MergeDispositionException(DestinationTerminalException):
    def __init__(
        self, dataset_name: str, staging_dataset_name: str, tables: Sequence[str], reason: str
    ) -> None:
        self.dataset_name = dataset_name
        self.staging_dataset_name = staging_dataset_name
        self.tables = tables
        self.reason = reason
        msg = (
            f"Merge sql job for `{dataset_name=:}` with `{staging_dataset_name=:}`"
            "COULD NOT BE GENERATED. Merge will not be performed. "
        )
        msg += (
            f"Data for the following tables `{tables}` is loaded to staging dataset. You may need"
            " to write your own materialization. The reason is:\n"
        )
        msg += reason
        super().__init__(msg)


class InvalidFilesystemLayout(DestinationTerminalException):
    def __init__(
        self,
        layout: str,
        expected_placeholders: Sequence[str],
        extra_placeholders: Sequence[str],
        invalid_placeholders: Sequence[str],
        unused_placeholders: Sequence[str],
    ) -> None:
        self.invalid_placeholders = invalid_placeholders
        self.extra_placeholders = extra_placeholders
        self.expected_placeholders = expected_placeholders
        self.unused_placeholders = unused_placeholders
        self.layout = layout

        message = (
            f"`{layout=:}` expected placeholders: `{expected_placeholders}` "
            f"Missing placeholders: `{invalid_placeholders}` "
        )

        if extra_placeholders:
            message += f"Extra placeholders specified: `{extra_placeholders}` "

        if unused_placeholders:
            message += f"Unused placeholders: `{unused_placeholders}`"

        super().__init__(message)


class InvalidPlaceholderCallback(DestinationTransientException):
    def __init__(self, callback_name: str) -> None:
        self.callback_name = callback_name
        super().__init__(
            f"Invalid placeholder `{callback_name=:}`, please make sure it can"
            " accept parameters the following `schema name`, `table name`,"
            " `load_id`, `file_id` and an `extension`",
        )


class CantExtractTablePrefix(DestinationTerminalException):
    def __init__(self, layout: str, details: str) -> None:
        msg = f"Cannot extract unique table prefix in `{layout=:}`. "
        msg += details
        msg += "An example of valid layout: `{table_name}/{load_id}.{file_id}.{ext}`"
        super().__init__(msg)



"""The Intro Pipeline Template contains the example from the docs intro page"""

# mypy: disable-error-code="no-untyped-def,arg-type"

from typing import Optional
import pandas as pd
import sqlalchemy as sa

import dlt
from dlt.sources.helpers import requests


def load_api_data() -> None:
    """Load data from the chess api, for more complex examples use our rest_api source"""

    # Create a dlt pipeline that will load
    # chess player data to the DuckDB destination
    pipeline = dlt.pipeline(
        pipeline_name="chess_pipeline", destination="duckdb", dataset_name="player_data"
    )
    # Grab some player data from Chess.com API
    data = []
    for player in ["magnuscarlsen", "rpragchess"]:
        response = requests.get(f"https://api.chess.com/pub/player/{player}")
        response.raise_for_status()
        data.append(response.json())

    # Extract, normalize, and load the data
    load_info = pipeline.run(data, table_name="player")
    print(load_info)  # noqa: T201


def load_pandas_data() -> None:
    """Load data from a public csv via pandas"""

    owid_disasters_csv = (
        "https://raw.githubusercontent.com/owid/owid-datasets/master/datasets/"
        "Natural%20disasters%20from%201900%20to%202019%20-%20EMDAT%20(2020)/"
        "Natural%20disasters%20from%201900%20to%202019%20-%20EMDAT%20(2020).csv"
    )
    df = pd.read_csv(owid_disasters_csv)

    pipeline = dlt.pipeline(
        pipeline_name="from_csv",
        destination="duckdb",
        dataset_name="mydata",
    )
    load_info = pipeline.run(df, table_name="natural_disasters")

    print(load_info)  # noqa: T201


def load_sql_data() -> None:
    """Load data from a sql database with sqlalchemy, for more complex examples use our sql_database source"""

    # Use any SQL database supported by SQLAlchemy, below we use a public
    # MySQL instance to get data.
    # NOTE: you'll need to install pymysql with `pip install pymysql`
    # NOTE: loading data from public mysql instance may take several seconds
    engine = sa.create_engine("mysql+pymysql://rfamro@mysql-rfam-public.ebi.ac.uk:4497/Rfam")

    with engine.connect() as conn:
        # Select genome table, stream data in batches of 100 elements
        query = "SELECT * FROM genome LIMIT 1000"
        rows = conn.execution_options(yield_per=100).exec_driver_sql(query)

        pipeline = dlt.pipeline(
            pipeline_name="from_database",
            destination="duckdb",
            dataset_name="genome_data",
        )

        # Convert the rows into dictionaries on the fly with a map function
        load_info = pipeline.run(map(lambda row: dict(row._mapping), rows), table_name="genome")

    print(load_info)  # noqa: T201


@dlt.resource(write_disposition="replace")
def github_api_resource(api_secret_key: Optional[str] = dlt.secrets.value):
    from dlt.sources.helpers.rest_client import paginate
    from dlt.sources.helpers.rest_client.auth import BearerTokenAuth
    from dlt.sources.helpers.rest_client.paginators import HeaderLinkPaginator

    url = "https://api.github.com/repos/dlt-hub/dlt/issues"

    # Github allows both authenticated and non-authenticated requests (with low rate limits)
    auth = BearerTokenAuth(api_secret_key) if api_secret_key else None
    for page in paginate(
        url, auth=auth, paginator=HeaderLinkPaginator(), params={"state": "open", "per_page": "100"}
    ):
        yield page


@dlt.source
def github_api_source(api_secret_key: Optional[str] = dlt.secrets.value):
    return github_api_resource(api_secret_key=api_secret_key)


def load_data_from_source():
    pipeline = dlt.pipeline(
        pipeline_name="github_api_pipeline", destination="duckdb", dataset_name="github_api_data"
    )
    load_info = pipeline.run(github_api_source())
    print(load_info)  # noqa: T201


if __name__ == "__main__":
    load_api_data()
    load_pandas_data()
    load_sql_data()



import os
import sys
import threading
import signal
from contextlib import contextmanager
from threading import Event
from types import FrameType
from typing import Any, Callable, Dict, Iterator, Optional, Union

from dlt.common.exceptions import SignalReceivedException

_received_signal: int = 0
exit_event = Event()
_signal_counts: Dict[int, int] = {}
_original_handlers: Dict[int, Union[int, Callable[[int, Optional[FrameType]], Any]]] = {}

# NOTE: do not use logger and print in signal handlers


def _signal_receiver(sig: int, frame: FrameType) -> None:
    """Handle POSIX signals with two-stage escalation.

    This handler is installed by intercepted_signals(). On the first occurrence of a
    supported signal (eg. SIGINT, SIGTERM) it requests a graceful shutdown by
    setting a process-wide flag and waking sleeping threads via exit_event.
    A second occurrence of the same signal escalates by delegating to the
    original handler or the system default, which typically results in an
    immediate process termination (eg. KeyboardInterrupt for SIGINT).

    Args:
        sig: Signal number (for example, signal.SIGINT or signal.SIGTERM).
        frame: The current stack frame when the signal was received.

    Notes:
        - The CPython runtime delivers signal handlers in the main thread only.
          Worker threads must cooperatively observe shutdown via raise_if_signalled()
          or the signal-aware sleep().
    """
    # track how many times this signal type has been received
    _signal_counts[sig] = _signal_counts.get(sig, 0) + 1

    if _signal_counts[sig] == 1:
        # first signal of this type: set flag and wake threads
        set_received_signal(sig)
        if sys.stdin.isatty():
            # log to console using low level functions that are safe for signal handlers
            if sig == signal.SIGINT:
                sig_desc = "CTRL-C"
            else:
                sig_desc = f"Signal {sig}"
            msg = (
                f"{sig_desc} received. Trying to shut down gracefully. It may take time to drain"
                f" job pools. Send {sig_desc} again to force stop.\n"
            )
            try:
                os.write(sys.stderr.fileno(), msg.encode(encoding="utf-8"))
            except OSError:
                pass
    elif _signal_counts[sig] >= 2:
        # second signal of this type: call original handler
        original_handler = _original_handlers.get(sig, signal.SIG_DFL)
        if callable(original_handler):
            original_handler(sig, frame)
        elif original_handler == signal.SIG_DFL:
            # restore default and re-raise to trigger default behavior
            signal.signal(sig, signal.SIG_DFL)
            signal.raise_signal(sig)

    exit_event.set()


def _clear_signals() -> None:
    global _received_signal

    _received_signal = 0
    _signal_counts.clear()
    _original_handlers.clear()


def set_received_signal(sig: int) -> None:
    """Called when signal was received"""
    global _received_signal

    _received_signal = sig


def raise_if_signalled() -> None:
    """Raises `SignalReceivedException` if signal was received."""
    if was_signal_received():
        raise exception_for_signal()


def exception_for_signal() -> BaseException:
    if not was_signal_received():
        raise RuntimeError("no signal received")
    return SignalReceivedException(_received_signal)


def was_signal_received() -> bool:
    """check if a signal was received"""
    return True if _received_signal else False


def sleep(sleep_seconds: float) -> None:
    """A signal-aware version of sleep function. Will wake up if signal is received but will not raise exception."""
    # sleep or wait for signal
    exit_event.clear()
    exit_event.wait(sleep_seconds)


def wake_all() -> None:
    """Wakes all threads sleeping on event"""
    exit_event.set()


@contextmanager
def intercepted_signals() -> Iterator[None]:
    """Will intercept SIGINT and SIGTERM and will delay calling signal handlers until
    `raise_if_signalled` is explicitly used or when a second signal with the same int value arrives.

    A no-op when not called on main thread.

    Can be nested - nested calls are no-ops.
    """

    if threading.current_thread() is threading.main_thread():
        # check if handlers are already installed (nested call)
        current_sigint_handler = signal.getsignal(signal.SIGINT)

        if current_sigint_handler is _signal_receiver:
            # already installed, this is a nested call - just yield
            yield
            return

        # First call - install handlers
        original_sigint_handler = current_sigint_handler
        original_sigterm_handler = signal.getsignal(signal.SIGTERM)

        # store original handlers for signal_receiver to use
        _original_handlers[signal.SIGINT] = original_sigint_handler
        _original_handlers[signal.SIGTERM] = original_sigterm_handler

        try:
            signal.signal(signal.SIGINT, _signal_receiver)
            signal.signal(signal.SIGTERM, _signal_receiver)
            yield
        finally:
            signal.signal(signal.SIGINT, original_sigint_handler)
            signal.signal(signal.SIGTERM, original_sigterm_handler)
            _clear_signals()

    else:
        from dlt.common import logger

        logger.info("Running in daemon thread, signals not enabled")
        yield



from .naming import NamingConvention as WeaviateNamingConvention


class NamingConvention(WeaviateNamingConvention):
    """Case insensitive naming convention for Weaviate. Lower cases all identifiers"""

    @property
    def is_case_sensitive(self) -> bool:
        return False

    def _lowercase_property(self, identifier: str) -> str:
        """Lowercase the whole property to become case insensitive"""
        return identifier.lower()



from dlt.common.exceptions import DltException


class RESTClientException(DltException):
    pass


class IgnoreResponseException(RESTClientException):
    pass


class PaginatorSetupError(RESTClientException, ValueError):
    pass


class PaginatorNotFound(RESTClientException):
    pass



from __future__ import annotations

from typing import IO, Any, Dict, Iterator, List, Sequence, TYPE_CHECKING, Optional
import math
from urllib.parse import quote_plus, urlencode

from dlt.common.libs.sql_alchemy import sa

from dlt.common import logger
from dlt.common.destination.client import (
    RunnableLoadJob,
    HasFollowupJobs,
    PreparedTableSchema,
)
from dlt.common.storages import FileStorage
from dlt.common.json import json, PY_DATETIME_DECODERS

from dlt.destinations._adbc_jobs import AdbcParquetCopyJob
from dlt.destinations.sql_jobs import SqlFollowupJob
from dlt.destinations.impl.sqlalchemy.db_api_client import SqlalchemyClient
from dlt.destinations.impl.sqlalchemy.merge_job import SqlalchemyMergeFollowupJob

if TYPE_CHECKING:
    from dlt.destinations.impl.sqlalchemy.sqlalchemy_job_client import SqlalchemyJobClient


def build_mysql_adbc_dsn(
    username: Optional[str] = None,
    password: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[int] = None,
    database: Optional[str] = None,
    params: Optional[Dict[str, str]] = None,
) -> str:
    """Build a DSN connection string for the go-mysql ADBC driver.

    The go-mysql driver (github.com/go-sql-driver/mysql) has specific DSN format:
        [username[:password]@][protocol[(address)]]/dbname[?param1=value1&...&paramN=valueN]

    Based on the driver's source code (dsn.go):
    - Username and password are NOT URL-decoded by the driver
    - Database name IS URL-decoded using url.PathUnescape
    - Query parameter values ARE URL-decoded using url.QueryUnescape

    Args:
        username: MySQL username (not URL-encoded)
        password: MySQL password (not URL-encoded)
        host: MySQL host
        port: MySQL port (default: 3306)
        database: Database/schema name (will be URL-encoded)
        params: Query parameters (values will be URL-encoded)

    Returns:
        DSN connection string for go-mysql driver
    """
    # Build auth part - username and password are NOT encoded per go-mysql driver behavior
    auth = ""
    if username or password:
        auth = f"{username or ''}:{password or ''}@"

    # Host and port
    host = host or "localhost"
    port = port or 3306

    # Database name is URL-encoded using PathEscape equivalent (quote_plus with safe='')
    # The go-mysql driver uses url.PathUnescape to decode it
    db_encoded = quote_plus(database or "", safe="")

    base = f"{auth}tcp({host}:{port})/{db_encoded}"

    # Query parameters - values are URL-encoded
    if params:
        conn_str = f"{base}?{urlencode(params)}"
    else:
        conn_str = base

    return conn_str


class SqlalchemyJsonLInsertJob(RunnableLoadJob, HasFollowupJobs):
    def __init__(self, file_path: str, table: sa.Table) -> None:
        super().__init__(file_path)
        self._job_client: "SqlalchemyJobClient" = None
        self.table = table

    def _open_load_file(self) -> IO[bytes]:
        return FileStorage.open_zipsafe_ro(self._file_path, "rb")

    def _iter_data_items(self) -> Iterator[Dict[str, Any]]:
        all_cols = {col.name: None for col in self.table.columns}
        with FileStorage.open_zipsafe_ro(self._file_path, "rb") as f:
            for line in f:
                # Decode date/time to py datetime objects. Some drivers have issues with pendulum objects
                for item in json.typed_loadb(line, decoders=PY_DATETIME_DECODERS):
                    # Fill any missing columns in item with None. Bulk insert fails when items have different keys
                    if item.keys() != all_cols.keys():
                        yield {**all_cols, **item}
                    else:
                        yield item

    def _iter_data_item_chunks(self) -> Iterator[Sequence[Dict[str, Any]]]:
        max_rows = self._job_client.capabilities.max_rows_per_insert or math.inf
        # Limit by max query length should not be needed,
        # bulk insert generates an INSERT template with a single VALUES tuple of placeholders
        # If any dialects don't do that we need to check the str length of the query
        # TODO: Max params may not be needed. Limits only apply to placeholders in sql string (mysql/sqlite)
        max_params = self._job_client.capabilities.max_query_parameters or math.inf
        chunk: List[Dict[str, Any]] = []
        params_count = 0
        for item in self._iter_data_items():
            if len(chunk) + 1 == max_rows or params_count + len(item) > max_params:
                # Rotate chunk
                yield chunk
                chunk = []
                params_count = 0
            params_count += len(item)
            chunk.append(item)

        if chunk:
            yield chunk

    def run(self) -> None:
        _sql_client = self._job_client.sql_client
        # Copy the table to the current dataset (i.e. staging) if needed
        # This is a no-op if the table is already in the correct schema
        table = self.table.to_metadata(
            self.table.metadata, schema=_sql_client.dataset_name  # type: ignore[attr-defined]
        )

        with _sql_client.begin_transaction():
            for chunk in self._iter_data_item_chunks():
                _sql_client.execute_sql(table.insert(), chunk)


class SqlalchemyParquetADBCJob(AdbcParquetCopyJob):
    """ADBC Parquet copy job for SQLAlchemy (sqlite, mysql) with query param handling."""

    def __init__(self, file_path: str, table: sa.Table) -> None:
        super().__init__(file_path)
        self._job_client: "SqlalchemyJobClient" = None
        self.table = table

    if TYPE_CHECKING:
        from adbc_driver_manager.dbapi import Connection

    def _connect(self) -> Connection:
        from adbc_driver_manager import dbapi

        engine = self._job_client.config.credentials.engine
        dialect = engine.dialect.name.lower()
        url = engine.url

        query = dict(url.query or {})

        if dialect == "sqlite":
            # disable schema and catalog when ingest
            self._connect_schema_name = ""
            self._connect_catalog_name = ""

            # attach directly to dataset sqlite file as "main"
            if self._job_client.sql_client.dataset_name == "main":
                db_path = url.database
            else:
                db_path = self._job_client.sql_client._sqlite_dataset_filename(
                    self._job_client.sql_client.dataset_name
                )
            conn_str = f"file:{db_path}"

            if query:
                conn_str = f"{conn_str}?{urlencode(query)}"

            conn = dbapi.connect(driver="sqlite", db_kwargs={"uri": conn_str})
            # WAL mode already set, add busy timeout to handle multiple writers
            with conn.cursor() as c:
                c.execute("PRAGMA busy_timeout=1000;")
            return conn

        elif dialect == "mysql":
            # disable schema and catalog when ingest
            self._connect_schema_name = ""
            self._connect_catalog_name = ""

            # mysql: convert SSL params into go-mysql ADBC parameters
            mapped = {}
            for k, v in query.items():
                lk = k.lower()
                if lk == "ssl_ca":
                    mapped["tls-ca"] = v
                elif lk == "ssl_cert":
                    mapped["tls-cert"] = v
                elif lk == "ssl_key":
                    mapped["tls-key"] = v
                elif lk == "ssl_mode":
                    mapped["tls"] = v
                else:
                    mapped[k] = v

            conn_str = build_mysql_adbc_dsn(
                username=url.username,
                password=str(url.password),
                host=url.host,
                port=url.port,
                database=self._job_client.sql_client.dataset_name,
                params=mapped,  # type: ignore
            )

            return dbapi.connect(driver="mysql", db_kwargs={"uri": conn_str})

        else:
            raise NotImplementedError(f"ADBC not supported for sqlalchemy dialect {dialect}")


class SqlalchemyParquetInsertJob(SqlalchemyJsonLInsertJob):
    def _iter_data_item_chunks(self) -> Iterator[Sequence[Dict[str, Any]]]:
        from dlt.common.libs.pyarrow import ParquetFile

        num_cols = len(self.table.columns)
        max_rows = self._job_client.capabilities.max_rows_per_insert or None
        max_params = self._job_client.capabilities.max_query_parameters or None
        read_limit = None

        with ParquetFile(self._file_path) as reader:
            if max_params is not None:
                read_limit = math.floor(max_params / num_cols)

            if max_rows is not None:
                if read_limit is None:
                    read_limit = max_rows
                else:
                    read_limit = min(read_limit, max_rows)

            if read_limit is None:
                yield reader.read().to_pylist()
                return

            for chunk in reader.iter_batches(batch_size=read_limit):
                yield chunk.to_pylist()


class SqlalchemyReplaceJob(SqlFollowupJob):
    @classmethod
    def generate_sql(
        cls,
        table_chain: Sequence[PreparedTableSchema],
        sql_client: SqlalchemyClient,  # type: ignore[override]
    ) -> List[str]:
        statements: List[str] = []
        for table in table_chain:
            # Tables must have already been created in metadata
            table_obj = sql_client.get_existing_table(table["name"])
            staging_table_obj = table_obj.to_metadata(
                sql_client.metadata, schema=sql_client.staging_dataset_name
            )
            stmt = str(table_obj.delete().compile(dialect=sql_client.dialect))
            statements.append(stmt)

            stmt = str(
                table_obj.insert()
                .from_select(
                    [col.name for col in staging_table_obj.columns], staging_table_obj.select()
                )
                .compile(dialect=sql_client.dialect)
            )

            statements.append(stmt)

        return statements


__all__ = [
    "SqlalchemyJsonLInsertJob",
    "SqlalchemyParquetInsertJob",
    "SqlalchemyReplaceJob",
    "SqlalchemyMergeFollowupJob",
]



from typing import TYPE_CHECKING, Dict, Iterator, Optional, Sequence, List, Any

from dlt.common import logger
from dlt.common.destination.client import (
    FollowupJobRequest,
    HasFollowupJobs,
    LoadJob,
    PreparedTableSchema,
    RunnableLoadJob,
)
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.schema import TColumnSchema, TColumnHint, Schema
from dlt.common.schema.typing import TColumnType

from dlt.common.storages.load_package import ParsedLoadJobFileName
from dlt.destinations._adbc_jobs import AdbcParquetCopyJob
from dlt.destinations.sql_jobs import SqlStagingReplaceFollowupJob, SqlMergeFollowupJob

from dlt.destinations.insert_job_client import InsertValuesJobClient

from dlt.destinations.impl.mssql.sql_client import PyOdbcMsSqlClient
from dlt.destinations.impl.mssql.configuration import MsSqlClientConfiguration, build_odbc_dsn
from dlt.destinations.sql_client import SqlClientBase


HINT_TO_MSSQL_ATTR: Dict[TColumnHint, str] = {"unique": "UNIQUE"}
VARCHAR_MAX_N: int = 4000
VARBINARY_MAX_N: int = 8000


class MsSqlStagingReplaceJob(SqlStagingReplaceFollowupJob):
    @classmethod
    def generate_sql(
        cls,
        table_chain: Sequence[PreparedTableSchema],
        sql_client: SqlClientBase[Any],
    ) -> List[str]:
        sql: List[str] = []
        for table in table_chain:
            with sql_client.with_staging_dataset():
                staging_table_name = sql_client.make_qualified_table_name(table["name"])
            table_name = sql_client.make_qualified_table_name(table["name"])
            # drop destination table
            sql.append(f"DROP TABLE IF EXISTS {table_name}")
            # moving staging table to destination schema
            sql.append(
                f"ALTER SCHEMA {sql_client.fully_qualified_dataset_name()} TRANSFER"
                f" {staging_table_name}"
            )
            # recreate staging table
            sql.append(f"SELECT * INTO {staging_table_name} FROM {table_name} WHERE 1 = 0")
        return sql


class MsSqlMergeJob(SqlMergeFollowupJob):
    @classmethod
    def gen_key_table_clauses(
        cls,
        root_table_name: str,
        staging_root_table_name: str,
        key_clauses: Sequence[str],
        for_delete: bool,
    ) -> List[str]:
        """Generate sql clauses that may be used to select or delete rows in root table of destination dataset"""
        if for_delete:
            # MS SQL doesn't support alias in DELETE FROM
            return [
                f"FROM {root_table_name} WHERE EXISTS (SELECT 1 FROM"
                f" {staging_root_table_name} WHERE"
                f" {' OR '.join([c.format(d=root_table_name,s=staging_root_table_name) for c in key_clauses])})"
            ]
        return SqlMergeFollowupJob.gen_key_table_clauses(
            root_table_name, staging_root_table_name, key_clauses, for_delete
        )

    @classmethod
    def _to_temp_table(cls, select_sql: str, temp_table_name: str, unique_column: str) -> str:
        return f"SELECT * INTO {temp_table_name} FROM ({select_sql}) as t"

    @classmethod
    def _new_temp_table_name(cls, table_name: str, op: str, sql_client: SqlClientBase[Any]) -> str:
        return SqlMergeFollowupJob._new_temp_table_name("#" + table_name, op, sql_client)


class MssqlParquetCopyJob(AdbcParquetCopyJob):
    _config: MsSqlClientConfiguration

    if TYPE_CHECKING:
        from adbc_driver_manager.dbapi import Connection

    def _connect(self) -> "Connection":
        from adbc_driver_manager import dbapi

        self._config = self._job_client.config  # type: ignore[assignment]
        conn_dsn = self.odbc_to_go_mssql_dsn(self._config.credentials.get_odbc_dsn_dict())
        conn_str = build_odbc_dsn(conn_dsn)
        return dbapi.connect(driver="mssql", db_kwargs={"uri": conn_str})

    @staticmethod
    def odbc_to_go_mssql_dsn(dsn: Dict[str, Any]) -> Dict[str, Any]:
        """Converts odbc connection string to go connection string used by ADBC"""
        # DSN keys are already normalized to upper case
        result: Dict[str, Any] = {}

        for upper, value in dsn.items():
            if value is None:
                continue

            v = str(value)

            if upper == "ENCRYPT":
                v = v.strip().lower()

                # ODBC: yes/mandatory/true/1 → go-mssqldb: true (TLS on)
                if v in {"yes", "true", "1", "mandatory"}:
                    v = "true"

                # ODBC: strict → go-mssqldb strict (if supported by the driver)
                elif v in {"strict"}:
                    v = "strict"

                # ODBC: optional → go-mssqldb optional (login only)
                elif v in {"optional"}:
                    v = "optional"

                # ODBC: no/false/0/disabled → go-mssqldb disable (no TLS at all)
                # This mirrors your previous string hack:
                #   .replace("=yes", "=1").replace("=no", "=disable")
                elif v in {"no", "false", "0", "disabled", "disable"}:
                    v = "disable"

            elif upper == "TRUSTSERVERCERTIFICATE":
                v = v.strip().lower()

                # ODBC uses yes/no; go-mssqldb expects true/false (but is lenient);
                # we normalize explicitly.
                if v in {"yes", "true", "1"}:
                    v = "true"
                elif v in {"no", "false", "0"}:
                    v = "false"

            result[upper] = v

        return result


class MsSqlJobClient(InsertValuesJobClient):
    def __init__(
        self,
        schema: Schema,
        config: MsSqlClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        dataset_name, staging_dataset_name = InsertValuesJobClient.create_dataset_names(
            schema, config
        )
        sql_client = PyOdbcMsSqlClient(
            dataset_name,
            staging_dataset_name,
            config.credentials,
            capabilities,
        )
        super().__init__(schema, config, sql_client)
        self.config: MsSqlClientConfiguration = config
        self.sql_client = sql_client
        self.active_hints = HINT_TO_MSSQL_ATTR if self.config.create_indexes else {}
        self.type_mapper = capabilities.get_type_mapper()

    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        job = super().create_load_job(table, file_path, load_id, restore)
        if not job:
            parsed_file = ParsedLoadJobFileName.parse(file_path)
            if parsed_file.file_format == "parquet":
                job = MssqlParquetCopyJob(file_path)
        return job

    def _create_merge_followup_jobs(
        self, table_chain: Sequence[PreparedTableSchema]
    ) -> List[FollowupJobRequest]:
        return [MsSqlMergeJob.from_table_chain(table_chain, self.sql_client)]

    def _make_add_column_sql(
        self, new_columns: Sequence[TColumnSchema], table: PreparedTableSchema = None
    ) -> List[str]:
        # Override because mssql requires multiple columns in a single ADD COLUMN clause
        return ["ADD \n" + ",\n".join(self._get_column_def_sql(c, table) for c in new_columns)]

    def _get_column_def_sql(self, c: TColumnSchema, table: PreparedTableSchema = None) -> str:
        sc_type = c["data_type"]
        if sc_type == "text" and c.get("unique"):
            # MSSQL does not allow index on large TEXT columns
            db_type = "nvarchar(%i)" % (c.get("precision") or 900)
        else:
            db_type = self.type_mapper.to_destination_type(c, table)

        hints_str = self._get_column_hints_sql(c)
        column_name = self.sql_client.escape_column_name(c["name"])
        return f"{column_name} {db_type} {hints_str} {self._gen_not_null(c.get('nullable', True))}"

    def _create_replace_followup_jobs(
        self, table_chain: Sequence[PreparedTableSchema]
    ) -> List[FollowupJobRequest]:
        root_table = table_chain[0]
        if root_table["x-replace-strategy"] == "staging-optimized":  # type: ignore[typeddict-item]
            return [MsSqlStagingReplaceJob.from_table_chain(table_chain, self.sql_client)]
        return super()._create_replace_followup_jobs(table_chain)

    def _from_db_type(
        self, pq_t: str, precision: Optional[int], scale: Optional[int]
    ) -> TColumnType:
        return self.type_mapper.from_destination_type(pq_t, precision, scale)



import dataclasses
from typing import Optional, Final, Callable, Union, Any
from typing_extensions import ParamSpec

from dlt.common.configuration import configspec, ConfigurationValueError
from dlt.common.destination import TLoaderFileFormat
from dlt.common.destination.client import (
    DestinationClientConfiguration,
)
from dlt.common.typing import TDataItems
from dlt.common.schema import TTableSchema

TDestinationCallable = Callable[[Union[TDataItems, str], TTableSchema], None]
TDestinationCallableParams = ParamSpec("TDestinationCallableParams")


def dummy_custom_destination(*args: Any, **kwargs: Any) -> None:
    pass


@configspec
class CustomDestinationClientConfiguration(DestinationClientConfiguration):
    destination_type: Final[str] = dataclasses.field(default="destination", init=False, repr=False, compare=False)  # type: ignore
    destination_callable: Optional[Union[str, TDestinationCallable]] = None  # noqa: A003
    loader_file_format: TLoaderFileFormat = "typed-jsonl"
    batch_size: int = 10
    skip_dlt_columns_and_tables: bool = True
    max_table_nesting: Optional[int] = 0

    def ensure_callable(self) -> None:
        """Makes sure that valid callable was provided"""
        # TODO: this surely can be done with `on_resolved`
        if (
            self.destination_callable is None
            or self.destination_callable is dummy_custom_destination
        ):
            raise ConfigurationValueError(
                f"A valid callable was not provided to `{self.__class__.__name__}`. Did you"
                " properly decorate the function with `@dlt.destination`?"
            )



from typing import Any, Optional, List
import os
from dlt.common.configuration.providers.toml import (
    ConfigTomlProvider,
    SecretsTomlProvider,
)


class ProfilePathMixin:
    def __init__(self, *args: Any, profile: str, **kwargs: Any) -> None:
        """A mixin that adds profile-aware path resolution by overriding `_resolve_toml_paths`"""
        self._profile = profile
        super().__init__(*args, **kwargs)

    def _resolve_toml_paths(self, file_name: str, resolvable_dirs: List[str]) -> List[str]:
        resolvable_files = []
        for d in resolvable_dirs:
            # append each a profile and a base file name for each directory
            # profile name is always first
            resolvable_files.append(os.path.join(d, f"{self._profile}.{file_name}"))
            resolvable_files.append(os.path.join(d, file_name))
        return resolvable_files


class ProfileSecretsTomlProvider(ProfilePathMixin, SecretsTomlProvider):
    def __init__(self, settings_dir: str, profile: str, global_dir: Optional[str] = None) -> None:
        """a secret toml provider loading from {profile}.secrets.toml file."""
        super().__init__(settings_dir=settings_dir, global_dir=global_dir, profile=profile)


class ProfileConfigTomlProvider(ProfilePathMixin, ConfigTomlProvider):
    def __init__(self, settings_dir: str, profile: str, global_dir: Optional[str] = None) -> None:
        """a config toml provider loading from {profile}.config.toml file."""
        super().__init__(settings_dir=settings_dir, global_dir=global_dir, profile=profile)



import re
from functools import lru_cache
from typing import ClassVar

from dlt.common.normalizers.naming.naming import NamingConvention as BaseNamingConvention
from dlt.common.normalizers.naming.sql_cs_v1 import (
    RE_UNDERSCORES,
    RE_LEADING_DIGITS,
    RE_NON_ALPHANUMERIC,
)
from dlt.common.typing import REPattern


class NamingConvention(BaseNamingConvention):
    """Case insensitive naming convention, converting source identifiers into lower case snake case with reduced alphabet.

    - Spaces around identifier are trimmed
    - Removes all ascii characters except ascii alphanumerics and underscores
    - Prepends `_` if name starts with number.
    - Multiples of `_` are converted into single `_`.
    - Replaces all trailing `_` with `x`
    - Replaces `+` and `*` with `x`, `-` with `_`, `@` with `a` and `|` with `l`

    Uses __ as parent-child separator for tables and flattened column names.
    """

    RE_UNDERSCORES: ClassVar[REPattern] = RE_UNDERSCORES
    RE_LEADING_DIGITS: ClassVar[REPattern] = RE_LEADING_DIGITS
    RE_NON_ALPHANUMERIC: ClassVar[REPattern] = RE_NON_ALPHANUMERIC

    _SNAKE_CASE_BREAK_1 = re.compile("([^_])([A-Z][a-z]+)")
    _SNAKE_CASE_BREAK_2 = re.compile("([a-z0-9])([A-Z])")
    _REDUCE_ALPHABET = ("+-*@|", "x_xal")
    _TR_REDUCE_ALPHABET = str.maketrans(_REDUCE_ALPHABET[0], _REDUCE_ALPHABET[1])

    @property
    def is_case_sensitive(self) -> bool:
        return False

    def normalize_identifier(self, identifier: str) -> str:
        identifier = super().normalize_identifier(identifier)
        # print(f"{identifier} -> {self.shorten_identifier(identifier, self.max_length)} ({self.max_length})")
        return self._normalize_identifier(identifier, self.max_length)

    @staticmethod
    @lru_cache(maxsize=None)
    def _normalize_identifier(identifier: str, max_length: int) -> str:
        """Normalizes the identifier according to naming convention represented by this function"""
        # all characters that are not letters digits or a few special chars are replaced with underscore
        normalized_ident = identifier.translate(NamingConvention._TR_REDUCE_ALPHABET)
        normalized_ident = NamingConvention.RE_NON_ALPHANUMERIC.sub("_", normalized_ident)

        # shorten identifier
        return NamingConvention.shorten_identifier(
            NamingConvention._to_snake_case(normalized_ident), identifier, max_length
        )

    @classmethod
    def _to_snake_case(cls, identifier: str) -> str:
        # then convert to snake case
        identifier = cls._SNAKE_CASE_BREAK_1.sub(r"\1_\2", identifier)
        identifier = cls._SNAKE_CASE_BREAK_2.sub(r"\1_\2", identifier).lower()

        # leading digits will be prefixed (if regex is defined)
        if cls.RE_LEADING_DIGITS and cls.RE_LEADING_DIGITS.match(identifier):
            identifier = "_" + identifier

        # replace trailing _ with x
        stripped_ident = identifier.rstrip("_")
        strip_count = len(identifier) - len(stripped_ident)
        stripped_ident += "x" * strip_count

        # identifier = cls._RE_ENDING_UNDERSCORES.sub("x", identifier)
        # replace consecutive underscores with single one to prevent name collisions with PATH_SEPARATOR
        return cls.RE_UNDERSCORES.sub("_", stripped_ident)



import inspect
import types
from typing import (
    AsyncIterator,
    ClassVar,
    Dict,
    Sequence,
    Union,
    Iterator,
    List,
    Awaitable,
    Tuple,
    Type,
    Literal,
)
from concurrent.futures import TimeoutError as FutureTimeoutError

from dlt.common.configuration import configspec
from dlt.common.configuration.inject import with_config
from dlt.common.configuration.specs import (
    BaseConfiguration,
    ContainerInjectableContext,
    known_sections,
)
from dlt.common.configuration.container import Container
from dlt.common.exceptions import PipelineException
from dlt.common.utils import get_callable_name

from dlt.extract.exceptions import (
    DltSourceException,
    ExtractorException,
    InvalidStepFunctionArguments,
    PipeException,
    PipeGenInvalid,
    PipeItemProcessingError,
    ResourceExtractionError,
)
from dlt.extract.pipe import Pipe
from dlt.extract.items import DataItemWithMeta, PipeItem, ResolvablePipeItem, SourcePipeItem
from dlt.extract.state import pipe_context
from dlt.extract.utils import wrap_async_iterator
from dlt.extract.concurrency import FuturesPool

TPipeNextItemMode = Literal["fifo", "round_robin"]


class PipeIterator(Iterator[PipeItem]):
    @configspec
    class PipeIteratorConfiguration(BaseConfiguration):
        max_parallel_items: int = 20
        workers: int = 5
        futures_poll_interval: float = 0.01
        copy_on_fork: bool = False
        next_item_mode: str = "round_robin"
        __section__: ClassVar[str] = known_sections.EXTRACT

    def __init__(
        self,
        max_parallel_items: int,
        workers: int,
        futures_poll_interval: float,
        sources: List[SourcePipeItem],
        next_item_mode: TPipeNextItemMode,
    ) -> None:
        self._sources = sources
        self._next_item_mode: TPipeNextItemMode = next_item_mode
        self._initial_sources_count = len(sources)
        self._current_source_index: int = 0
        self._futures_pool = FuturesPool(
            workers=workers,
            poll_interval=futures_poll_interval,
            max_parallel_items=max_parallel_items,
        )

    @classmethod
    @with_config(spec=PipeIteratorConfiguration)
    def from_pipe(
        cls,
        pipe: Pipe,
        *,
        max_parallel_items: int = 20,
        workers: int = 5,
        futures_poll_interval: float = 0.01,
        next_item_mode: TPipeNextItemMode = "round_robin",
    ) -> "PipeIterator":
        # join all dependent pipes
        if pipe.parent:
            pipe = pipe.full_pipe()
        # clone pipe to allow multiple iterations on pipe based on iterables/callables
        pipe = pipe._clone()
        # head must be iterator
        pipe.evaluate_gen()
        if not isinstance(pipe.gen, Iterator):
            raise PipeGenInvalid(pipe.name, pipe.gen)

        # create extractor
        sources = [SourcePipeItem(pipe.gen, 0, pipe, None)]
        return cls(max_parallel_items, workers, futures_poll_interval, sources, next_item_mode)

    @classmethod
    @with_config(spec=PipeIteratorConfiguration)
    def from_pipes(
        cls,
        pipes: Sequence[Pipe],
        yield_parents: bool = True,
        *,
        max_parallel_items: int = 20,
        workers: int = 5,
        futures_poll_interval: float = 0.01,
        copy_on_fork: bool = False,
        next_item_mode: TPipeNextItemMode = "round_robin",
    ) -> "PipeIterator":
        # print(f"max_parallel_items: {max_parallel_items} workers: {workers}")
        sources: List[SourcePipeItem] = []

        # clone all pipes before iterating (recursively) as we will fork them (this add steps) and evaluate gens
        pipes, _ = PipeIterator.clone_pipes(pipes)

        def _fork_pipeline(pipe: Pipe) -> None:
            if pipe.parent:
                # fork the parent pipe
                pipe.evaluate_gen()
                pipe.parent.fork(pipe, copy_on_fork=copy_on_fork)
                # make the parent yield by sending a clone of item to itself with position at the end
                if yield_parents and pipe.parent in pipes:
                    # fork is last step of the pipe so it will yield
                    pipe.parent.fork(pipe.parent, len(pipe.parent) - 1, copy_on_fork=copy_on_fork)
                _fork_pipeline(pipe.parent)
            else:
                # head of independent pipe must be iterator
                pipe.evaluate_gen()
                if not isinstance(pipe.gen, Iterator):
                    raise PipeGenInvalid(pipe.name, pipe.gen)
                # add every head as source only once
                if not any(i.pipe == pipe for i in sources):
                    sources.append(SourcePipeItem(pipe.gen, 0, pipe, None))

        # reverse pipes for current mode, as we start processing from the back
        pipes.reverse()
        for pipe in pipes:
            _fork_pipeline(pipe)

        # create extractor
        return cls(max_parallel_items, workers, futures_poll_interval, sources, next_item_mode)

    def __next__(self) -> PipeItem:
        pipe_item: Union[ResolvablePipeItem, SourcePipeItem] = None
        # __next__ should call itself to remove the `while` loop and continue clauses but that may lead to stack overflows: there's no tail recursion opt in python
        # https://stackoverflow.com/questions/13591970/does-python-optimize-tail-recursion (see Y combinator on how it could be emulated)
        while True:
            # do we need new item?
            if pipe_item is None:
                # Always check for done futures to avoid starving the pool
                pipe_item = self._futures_pool.resolve_next_future_no_wait()

                if pipe_item is None:
                    # if none then take element from the newest source
                    pipe_item = self._get_source_item()

                if pipe_item is None:
                    # Wait for some time for futures to resolve
                    try:
                        pipe_item = self._futures_pool.resolve_next_future(
                            use_configured_timeout=True
                        )
                    except FutureTimeoutError:
                        pass
                    else:
                        if pipe_item is None:
                            # pool was empty - then do a regular poll sleep
                            self._futures_pool.sleep()

                if pipe_item is None:
                    if self._futures_pool.empty and len(self._sources) == 0:
                        # no more elements in futures or sources
                        raise StopIteration()
                    else:
                        continue

            item = pipe_item.item
            # if item is iterator, then add it as a new source
            if isinstance(item, Iterator):
                self._sources.append(
                    SourcePipeItem(item, pipe_item.step, pipe_item.pipe, pipe_item.meta)
                )
                pipe_item = None
                continue

            # handle async iterator items as new source
            if isinstance(item, AsyncIterator):
                self._sources.append(
                    SourcePipeItem(
                        wrap_async_iterator(item), pipe_item.step, pipe_item.pipe, pipe_item.meta
                    ),
                )
                pipe_item = None
                continue

            with pipe_context(pipe_item.pipe):
                if isinstance(item, Awaitable) or callable(item):
                    # Callables are submitted to the pool to be executed in the background
                    self._futures_pool.submit(pipe_item)  # type: ignore[arg-type]
                    pipe_item = None
                    # Future will be resolved later, move on to the next item
                    continue

                # if we are at the end of the pipe then yield element
                if pipe_item.step == len(pipe_item.pipe) - 1:
                    # must be resolved
                    if isinstance(item, (Iterator, Awaitable, AsyncIterator)) or callable(item):
                        raise PipeItemProcessingError(
                            pipe_item.pipe.name,
                            f"Pipe item of type `{type(pipe_item.item).__name__}` was not fully"
                            f" evaluated at step `{pipe_item.step}`. This is an internal error or"
                            " you're yielding unexpected object from resources (e.g., fucntions,"
                            " awaitables).",
                        )
                    # mypy not able to figure out that item was resolved
                    return pipe_item  # type: ignore

                # advance to next step
                step = pipe_item.pipe[pipe_item.step + 1]
                try:
                    next_meta = pipe_item.meta
                    next_item = step(item, meta=pipe_item.meta)  # type: ignore
                    if isinstance(next_item, DataItemWithMeta):
                        next_meta = next_item.meta
                        next_item = next_item.data
                except TypeError as ty_ex:
                    assert callable(step)
                    raise InvalidStepFunctionArguments(
                        pipe_item.pipe.name,
                        get_callable_name(step),
                        inspect.signature(step),
                        str(ty_ex),
                    )
                except (PipelineException, ExtractorException, DltSourceException, PipeException):
                    raise
                except Exception as ex:
                    raise ResourceExtractionError(
                        pipe_item.pipe.name, step, str(ex), "transform"
                    ) from ex

            # create next pipe item if a value was returned. A None means that item was consumed/filtered out and should not be further processed
            if next_item is not None:
                pipe_item = ResolvablePipeItem(
                    next_item, pipe_item.step + 1, pipe_item.pipe, next_meta
                )
            else:
                pipe_item = None

    def _get_source_item(self) -> ResolvablePipeItem:
        sources_count = len(self._sources)
        # no more sources to iterate
        if sources_count == 0:
            return None
        try:
            first_evaluated_index: int = None
            # always reset to end of list for fifo mode, also take into account that new sources can be added
            # if too many new sources is added we switch to fifo not to exhaust them
            if self._next_item_mode == "fifo" or (
                sources_count - self._initial_sources_count >= self._futures_pool.max_parallel_items
            ):
                self._current_source_index = sources_count - 1
            else:
                self._current_source_index = (self._current_source_index - 1) % sources_count
            while True:
                # if we have checked all sources once and all returned None, return and poll/resolve some futures
                if self._current_source_index == first_evaluated_index:
                    return None
                # get next item from the current source
                gen, step, pipe, meta = self._sources[self._current_source_index]
                with pipe_context(pipe):
                    pipe_item = next(gen)
                if pipe_item is not None:
                    # full pipe item may be returned, this is used by ForkPipe step
                    # to redirect execution of an item to another pipe
                    # else
                    if not isinstance(pipe_item, ResolvablePipeItem):
                        # keep the item assigned step and pipe when creating resolvable item
                        if isinstance(pipe_item, DataItemWithMeta):
                            return ResolvablePipeItem(pipe_item.data, step, pipe, pipe_item.meta)
                        else:
                            return ResolvablePipeItem(pipe_item, step, pipe, meta)

                if pipe_item is not None:
                    return pipe_item

                # remember the first evaluated index
                if first_evaluated_index is None:
                    first_evaluated_index = self._current_source_index
                # always go round robin if None was returned or item is to be run as future
                self._current_source_index = (self._current_source_index - 1) % sources_count
        except StopIteration:
            # remove empty iterator and try another source
            self._sources.pop(self._current_source_index)
            # decrease initial source count if we popped an initial source
            if self._current_source_index < self._initial_sources_count:
                self._initial_sources_count -= 1
            return self._get_source_item()
        except (PipelineException, ExtractorException, DltSourceException, PipeException):
            raise
        except Exception as ex:
            raise ResourceExtractionError(pipe.name, gen, str(ex), "generator") from ex

    def close(self) -> None:
        # Close the futures pool and cancel all tasks
        # It's important to do this before closing generators as we can't close a running generator
        self._futures_pool.close()

        # close all generators
        for gen, _, _, _ in self._sources:
            if inspect.isgenerator(gen):
                gen.close()

        self._sources.clear()

    def __enter__(self) -> "PipeIterator":
        return self

    def __exit__(
        self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: types.TracebackType
    ) -> None:
        self.close()

    @staticmethod
    def clone_pipes(
        pipes: Sequence[Pipe], existing_cloned_pairs: Dict[str, Pipe] = None
    ) -> Tuple[List[Pipe], Dict[str, Pipe]]:
        """Clones pipes with their parents and fixes parent instances using pipe `instance_id`

        Pipe with given `instance_id` is cloned only once and then reused. This creates
        a tree of pipes that requires given pipe to be extracted only once, even if it
        has many child pipes that take data from it.
        """
        cloned_pipes = [
            p._clone() for p in pipes if p.instance_id not in (existing_cloned_pairs or {})
        ]
        cloned_pairs = {p.instance_id: c for p, c in zip(pipes, cloned_pipes)}
        if existing_cloned_pairs:
            cloned_pairs.update(existing_cloned_pairs)

        for clone in cloned_pipes:
            while True:
                if not clone.parent:
                    break
                # if already a clone
                if clone.parent in cloned_pairs.values():
                    break
                # clone if parent pipe not yet cloned
                parent_id = clone.parent.instance_id
                if parent_id not in cloned_pairs:
                    # print("cloning:" + clone.parent.name)
                    cloned_pairs[parent_id] = clone.parent._clone()
                # replace with clone
                # print(f"replace depends on {clone.name} to {clone.parent.name}")
                clone.parent = cloned_pairs[parent_id]
                # recur with clone
                clone = clone.parent

        return cloned_pipes, cloned_pairs


class ManagedPipeIterator(PipeIterator):
    """A version of the pipe iterator that gets closed automatically on an exception in _next_"""

    _ctx: List[ContainerInjectableContext] = None
    _container: Container = None

    def set_context(self, ctx: List[ContainerInjectableContext]) -> None:
        """Sets list of injectable contexts that will be injected into Container for each call to __next__"""
        self._ctx = ctx
        self._container = Container()

    def __next__(self) -> PipeItem:
        if self._ctx:
            managers = [self._container.injectable_context(ctx) for ctx in self._ctx]
            for manager in managers:
                manager.__enter__()
        try:
            item = super().__next__()
        except Exception as ex:
            # release context manager
            if self._ctx:
                if isinstance(ex, StopIteration):
                    for manager in managers:
                        manager.__exit__(None, None, None)
                else:
                    for manager in managers:
                        manager.__exit__(type(ex), ex, None)
            # crash in next
            self.close()
            raise
        if self._ctx:
            for manager in managers:
                manager.__exit__(None, None, None)
        return item



import os
import contextlib
from dlt.common.destination.client import DestinationClientDwhConfiguration
from dlt.common.destination.exceptions import DatasetNotAvailable
from dlt.common.exceptions import VenvNotFound
from dlt.common.runners import Venv
from dlt.common.schema import Schema
from dlt.common.typing import ConfigValue, TSecretStrValue
from dlt.common.schema.utils import normalize_schema_name

from dlt.helpers.dbt import (
    create_venv as _create_venv,
    package_runner as _package_runner,
    DBTPackageRunner,
    DEFAULT_DBT_VERSION as _DEFAULT_DBT_VERSION,
    restore_venv as _restore_venv,
)
from dlt.pipeline.pipeline import Pipeline


# TODO: make `venv_path` to be field in DBTRunnerConfiguration
def get_venv(
    pipeline: Pipeline, venv_path: str = "dbt", dbt_version: str = _DEFAULT_DBT_VERSION
) -> Venv:
    """Creates or restores a virtual environment in which the `dbt` packages are executed.

    The recommended way to execute dbt package is to use a separate virtual environment where only the dbt-core
    and required destination dependencies are installed. This avoid dependency clashes with the user-installed libraries.
    This method will create such environment at the location specified in `venv_path` and automatically install required dependencies
    as required by `pipeline`.

    Args:
        pipeline (Pipeline): A pipeline for which the required dbt dependencies are inferred
        venv_path (str, optional): A path where virtual environment is created or restored from.
                                   If relative path is provided, the environment will be created within pipeline's working directory.
                                   If path is None or empty, current venv will be used as is (no deps are installed)
                                   Defaults to "dbt".
        dbt_version (str, optional): Version of dbt to be used. Exact version (ie. "1.2.4") or pip requirements string (ie. ">=1.1<1.5" may be provided).

    Returns:
        Venv: A Virtual Environment with dbt dependencies installed
    """
    # use current environment if no path specified
    if not venv_path:
        return Venv.restore_current()

    # keep venv inside pipeline if path is relative
    if not os.path.isabs(venv_path):
        pipeline._pipeline_storage.create_folder(venv_path, exists_ok=True)
        venv_dir = pipeline._pipeline_storage.make_full_path_safe(venv_path)
    else:
        venv_dir = venv_path
    # try to restore existing venv
    with contextlib.suppress(VenvNotFound):
        # TODO: check dlt version in venv and update it if local version updated
        return _restore_venv(venv_dir, [pipeline.destination.spec().destination_type], dbt_version)

    return _create_venv(venv_dir, [pipeline.destination.spec().destination_type], dbt_version)


def package(
    pipeline: Pipeline,
    package_location: str,
    package_repository_branch: str = ConfigValue,
    package_repository_ssh_key: TSecretStrValue = "",
    auto_full_refresh_when_out_of_sync: bool = ConfigValue,
    venv: Venv = None,
) -> DBTPackageRunner:
    """Creates a Python wrapper over `dbt` package present at specified location, that allows to control it (ie. run and test) from Python code.

    The created wrapper minimizes the required effort to run `dbt` packages on datasets created with `dlt`. It clones the package repo and keeps it up to data,
    shares the `dlt` destination credentials with `dbt` and allows the isolated execution with `venv` parameter.
    The wrapper creates a `dbt` profile from `dlt` pipeline configuration. Specifically:
    1. destination is used to infer correct dbt profile
    2. destinations credentials are passed to dbt via environment variables
    3. dataset_name is used to configure the dbt database schema

    Args:
        pipeline (Pipeline): A pipeline containing destination, credentials and dataset_name used to configure the dbt package.
        package_location (str): A git repository url to be cloned or a local path where dbt package is present
        package_repository_branch (str, optional): A branch name, tag name or commit-id to check out. Defaults to None.
        package_repository_ssh_key (TSecretValue, optional): SSH key to be used to clone private repositories. Defaults to TSecretValue("").
        auto_full_refresh_when_out_of_sync (bool, optional): If set to True (default), the wrapper will automatically fall back to full-refresh mode when schema is out of sync
                                                             See: https://docs.getdbt.com/docs/build/incremental-models#what-if-the-columns-of-my-incremental-model-change
        venv (Venv, optional): A virtual environment with required dbt dependencies. Defaults to None which will execute dbt package in current environment.

    Returns:
        DBTPackageRunner: A configured and authenticated Python `dbt` wrapper
    """
    schema = (
        pipeline.default_schema
        if pipeline.default_schema_name
        else Schema(normalize_schema_name(pipeline.dataset_name or pipeline.pipeline_name))
    )
    job_client = pipeline._get_destination_clients(schema)[0]
    # check if config supports datasets
    if not isinstance(job_client.config, DestinationClientDwhConfiguration):
        raise DatasetNotAvailable(pipeline.destination.destination_description)
    if not venv:
        venv = Venv.restore_current()
    return _package_runner(
        venv,
        job_client.config,
        pipeline.working_dir,
        package_location,
        package_repository_branch,
        package_repository_ssh_key,
        auto_full_refresh_when_out_of_sync,
    )



import requests


def send_slack_message(incoming_hook: str, message: str, is_markdown: bool = True) -> None:
    from dlt.common import logger
    from dlt.common.json import json

    """Sends a `message` to  Slack `incoming_hook`, by default formatted as markdown."""
    r = requests.post(
        incoming_hook,
        data=json.dumps({"text": message, "mrkdwn": is_markdown}).encode("utf-8"),
        headers={"Content-Type": "application/json;charset=utf-8"},
    )
    if r.status_code >= 400:
        logger.warning(f"Could not post the notification to slack: {r.status_code}")
    r.raise_for_status()



import warnings
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional
from urllib.parse import urljoin, urlparse

from requests import Request, Response

from dlt.common import jsonpath
from dlt.common.utils import str2bool


class BasePaginator(ABC):
    """A base class for all paginator implementations. Paginators are used
    to handle paginated responses from RESTful APIs.

    See `RESTClient.paginate()` for example usage.
    """

    def __init__(self) -> None:
        self._has_next_page = True

    @property
    def has_next_page(self) -> bool:
        """Determines if there is a next page available.

        Returns:
            bool: True if a next page is available, otherwise False.
        """
        return self._has_next_page

    def init_request(self, request: Request) -> None:  # noqa: B027, optional override
        """Initializes the request object with parameters for the first
        pagination request.

        This method can be overridden by subclasses to include specific
        initialization logic.

        Args:
            request (Request): The request object to be initialized.
        """
        pass

    @abstractmethod
    def update_state(self, response: Response, data: Optional[List[Any]] = None) -> None:
        """Updates the paginator's state based on the response from the API.

        This method should extract necessary pagination details (like next page
        references) from the response and update the paginator's state
        accordingly. It should also set the `_has_next_page` attribute to
        indicate if there is a next page available.

        Args:
            response (Response): The response object from the API request.
        """
        ...

    @abstractmethod
    def update_request(self, request: Request) -> None:
        """Updates the request object with arguments for fetching the next page.

        This method should modify the request object to include necessary
        details (like URLs or parameters) for requesting the next page based on
        the current state of the paginator.

        Args:
            request (Request): The request object to be updated for the next
                page fetch.
        """
        ...

    def __str__(self) -> str:
        return f"{type(self).__name__} at {id(self):x}"


class SinglePagePaginator(BasePaginator):
    """A paginator for single-page API responses."""

    def update_state(self, response: Response, data: Optional[List[Any]] = None) -> None:
        self._has_next_page = False

    def update_request(self, request: Request) -> None:
        return


class RangePaginator(BasePaginator):
    """A base paginator class for paginators that use a numeric parameter
    for pagination, such as page number or offset.

    See `PageNumberPaginator` and `OffsetPaginator` for examples.
    """

    def __init__(
        self,
        param_name: Optional[str],
        initial_value: int,
        value_step: int,
        base_index: int = 0,
        maximum_value: Optional[int] = None,
        total_path: Optional[jsonpath.TJsonPath] = None,
        error_message_items: str = "items",
        stop_after_empty_page: Optional[bool] = True,
        *,
        has_more_path: Optional[jsonpath.TJsonPath] = None,
        param_body_path: Optional[str] = None,
    ):
        """
        Args:
            param_name (str): The query parameter name for the numeric value.
                For example, 'page'.
            param_body_path (str): The dot-separated path specifying
                where to place the numeric parameter in the request JSON body.
                Defaults to `None`. Either `param_name` or `param_body_path`
                must be provided, not both.
            initial_value (int): The initial value of the numeric parameter.
            value_step (int): The step size to increment the numeric parameter.
            base_index (int, optional): The index of the initial element.
                Used to define 0-based or 1-based indexing. Defaults to 0.
            maximum_value (int, optional): The maximum value for the numeric parameter.
                If provided, pagination will stop once this value is reached
                or exceeded, even if more data is available. This allows you
                to limit the maximum range for pagination.
                If not provided, `total_path` must be specified. Defaults to None.
            total_path (jsonpath.TJsonPath, optional): The JSONPath expression
                for the total number of items. For example, if the JSON response is
                `{"items": [...], "total": 100}`, the `total_path` would be 'total'.
                If not provided, `maximum_value` must be specified.
            error_message_items (str): The name of the items in the error message.
                Defaults to 'items'.
            stop_after_empty_page (bool): Whether pagination should stop when
              a page contains no result items. Defaults to `True`.
            has_more_path (jsonpath.TJsonPath): The JSONPath expression for
                the boolean value indicating whether there are more items to fetch.
                Defaults to None.
        """
        super().__init__()
        if (
            total_path is None
            and maximum_value is None
            and has_more_path is None
            and not stop_after_empty_page
        ):
            raise ValueError(
                "One of `total_path`, `maximum_value`, `has_more_path`, or `stop_after_empty_page`"
                " must be provided."
            )

        if bool(param_name) == bool(param_body_path):
            raise ValueError("Either 'param_name' or 'param_body_path' must be provided, not both.")

        self.param_name = param_name
        self.param_body_path = param_body_path
        self.initial_value = initial_value
        self.current_value = initial_value
        self.value_step = value_step
        self.base_index = base_index
        self.maximum_value = maximum_value
        self.total_path = jsonpath.compile_path(total_path) if total_path else None
        self.error_message_items = error_message_items
        self.stop_after_empty_page = stop_after_empty_page
        self.has_more_path = jsonpath.compile_path(has_more_path) if has_more_path else None

    def init_request(self, request: Request) -> None:
        self._has_next_page = True
        self.current_value = self.initial_value
        self.update_request(request)

    def update_state(self, response: Response, data: Optional[List[Any]] = None) -> None:
        if self._stop_after_this_page(data):
            self._has_next_page = False
        else:
            total = None
            if self.total_path:
                response_json = response.json()
                values = jsonpath.find_values(self.total_path, response_json)
                total = values[0] if values else None
                if total is None:
                    self._handle_missing_total(response_json)

                try:
                    total = int(total)
                except ValueError:
                    self._handle_invalid_total(total)

            self.current_value += self.value_step

            if (total is not None and self.current_value >= total + self.base_index) or (
                self.maximum_value is not None and self.current_value >= self.maximum_value
            ):
                self._has_next_page = False

            has_more = None
            if self.has_more_path:
                values = jsonpath.find_values(self.has_more_path, response.json())
                has_more = values[0] if values else None
                if has_more is None:
                    self._handle_missing_has_more(response.json())
                elif isinstance(has_more, str):
                    try:
                        has_more = str2bool(has_more)
                    except ValueError:
                        self._handle_invalid_has_more(has_more)
                elif not isinstance(has_more, bool):
                    self._handle_invalid_has_more(has_more)

                self._has_next_page = has_more

    def _stop_after_this_page(self, data: Optional[List[Any]] = None) -> bool:
        return self.stop_after_empty_page and not data

    def _handle_missing_total(self, response_json: Dict[str, Any]) -> None:
        raise ValueError(
            f"Total `{self.error_message_items}` not found in the response in"
            f" `{self.__class__.__name__}` .Expected a response with a `{self.total_path}` key, got"
            f" `{response_json}`."
        )

    def _handle_invalid_total(self, total: Any) -> None:
        raise ValueError(
            f"'{self.total_path}' is not an `int` in the response in `{self.__class__.__name__}`."
            f" Expected an integer, got `{total}`"
        )

    def _handle_missing_has_more(self, response_json: Dict[str, Any]) -> None:
        raise ValueError(
            f"Has more value not found in the response in `{self.__class__.__name__}`."
            f"Expected a response with a `{self.has_more_path}` key, got"
            f" `{response_json}`."
        )

    def _handle_invalid_has_more(self, has_more: Any) -> None:
        raise ValueError(
            f"'{self.has_more_path}' is not a `bool` in the response in"
            f" `{self.__class__.__name__}`. Expected a boolean, got `{has_more}`"
        )

    def update_request(self, request: Request) -> None:
        """Updates the request with the current value either in query parameters
        or in the request JSON body."""
        if self.param_body_path:
            self._update_request_with_body_path(request, self.param_body_path, self.current_value)
        else:
            self._update_request_with_param_name(request, self.param_name, self.current_value)

    @staticmethod
    def _update_request_with_param_name(
        request: Request, param_name: str, current_value: int
    ) -> None:
        if not param_name:
            raise ValueError("`param_name` must not be empty.")
        if request.params is None:
            request.params = {}
        request.params[param_name] = current_value

    @staticmethod
    def _update_request_with_body_path(
        request: Request, body_path: str, current_value: int
    ) -> None:
        if not body_path:
            raise ValueError("`body_path` must not be empty.")
        if request.json is None:
            request.json = {}
        jsonpath.set_value_at_path(request.json, body_path, current_value)


class PageNumberPaginator(RangePaginator):
    """A paginator that uses page number-based pagination strategy.

    For example, consider an API located at `https://api.example.com/items`
    that supports pagination through page number and page size query parameters,
    and provides the total number of pages in its responses, as shown below:

        {
            "items": [...],
            "total_pages": 10
        }

    To use `PageNumberPaginator` with such an API, you can instantiate `RESTClient`
    as follows:

        from dlt.sources.helpers.rest_client import RESTClient

        client = RESTClient(
            base_url="https://api.example.com",
            paginator=PageNumberPaginator(
                total_path="total_pages"
            )
        )

        @dlt.resource
        def get_items():
            for page in client.paginate("/items", params={"size": 100}):
                yield page

    Note that we pass the `size` parameter in the initial request to the API.
    The `PageNumberPaginator` will automatically increment the page number for
    each subsequent request until all items are fetched.

    If the API does not provide the total number of pages, you can use the
    `maximum_page` parameter to limit the number of pages to fetch. For example:

        client = RESTClient(
            base_url="https://api.example.com",
            paginator=PageNumberPaginator(
                maximum_page=5,
                total_path=None
            )
        )
        ...

    In this case, pagination will stop after fetching 5 pages of data.

    If the API provides a boolean value indicating whether there are more items
    to fetch, you can use the `has_more_path` parameter to stop the pagination
    when there are no more items to fetch. For example:

        {
            "items": [...],
            "has_more": false
        }

        client = RESTClient(
            base_url="https://api.example.com",
            paginator=PageNumberPaginator(
                total_path=None,
                has_more_path="has_more"
            )
        )
        ...

    In this case, pagination will stop once the `has_more` value is `false`.
    """

    def __init__(
        self,
        base_page: int = 0,
        page: int = None,
        page_param: Optional[str] = None,
        total_path: Optional[jsonpath.TJsonPath] = "total",
        maximum_page: Optional[int] = None,
        stop_after_empty_page: Optional[bool] = True,
        *,
        has_more_path: Optional[jsonpath.TJsonPath] = None,
        page_body_path: Optional[str] = None,
    ):
        """
        Args:
            base_page (int): The index of the initial page from the API perspective.
                Determines the page number that the API server uses for the starting
                page. Normally, this is 0-based or 1-based (e.g., 1, 2, 3, ...)
                indexing for the pages. Defaults to 0.
            page (int): The page number for the first request. If not provided,
                the initial value will be set to `base_page`.
            page_param (str): The query parameter name for the page number.
                Defaults to 'page'. Either `page_param` or `page_body_path`
                must be provided, not both.
            page_body_path (str): A dot-separated path specifying where
                to place the page number in the request JSON body. Use this instead
                of `page_param` when sending the page number in the request body.
                Defaults to `None`.
            total_path (jsonpath.TJsonPath): The JSONPath expression for
                the total number of pages. Defaults to 'total'.
            maximum_page (int): The maximum page number. If provided, pagination
                will stop once this page is reached or exceeded, even if more
                data is available. This allows you to limit the maximum number
                of pages for pagination. Defaults to None.
            stop_after_empty_page (bool): Whether pagination should stop when
              a page contains no result items. Defaults to `True`.
            has_more_path (jsonpath.TJsonPath): The JSONPath expression for
                the boolean value indicating whether there are more items to fetch.
                Defaults to None.
        """
        # For backward compatibility: set default cursor_param if both are None
        if page_param is None and page_body_path is None:
            page_param = "page"
        # Check that only one of page_param or page_body_path is provided
        elif (not page_param and not page_body_path) or (page_param and page_body_path):
            raise ValueError("Either 'page_param' or 'page_body_path' must be provided, not both.")

        if (
            total_path is None
            and maximum_page is None
            and has_more_path is None
            and not stop_after_empty_page
        ):
            raise ValueError(
                "One of `total_path`, `maximum_page`, `has_more_path`, or `stop_after_empty_page`"
                " must be provided."
            )

        page = page if page is not None else base_page

        super().__init__(
            param_name=page_param,
            param_body_path=page_body_path,
            initial_value=page,
            base_index=base_page,
            total_path=total_path,
            value_step=1,
            maximum_value=maximum_page,
            error_message_items="pages",
            stop_after_empty_page=stop_after_empty_page,
            has_more_path=has_more_path,
        )

    def __str__(self) -> str:
        return (
            super().__str__()
            + f": current page: {self.current_value} "
            + (f"page_param: {self.param_name} " if self.param_name else "")
            + (f"page_body_path: {self.param_body_path} " if self.param_body_path else "")
            + f"total_path: {self.total_path} maximum_value: {self.maximum_value} "
            f"has_more_path: {self.has_more_path}"
        )


class OffsetPaginator(RangePaginator):
    """A paginator that uses offset-based pagination strategy.

    This paginator is useful for APIs where pagination is controlled
    through offset and limit query parameters and the total count of items
    is returned in the response.

    For example, consider an API located at `https://api.example.com/items`
    that supports pagination through offset and limit, and provides the total
    item count in its responses, as shown below:

        {
            "items": [...],
            "total": 1000
        }

    To use `OffsetPaginator` with such an API, you can instantiate `RESTClient`
    as follows:

        from dlt.sources.helpers.rest_client import RESTClient

        client = RESTClient(
            base_url="https://api.example.com",
            paginator=OffsetPaginator(
                limit=100,
                total_path="total"
            )
        )
        @dlt.resource
        def get_items():
            for page in client.paginate("/items"):
                yield page

    The `OffsetPaginator` will automatically increment the offset for each
    subsequent request until all items are fetched.

    If the API does not provide the total count of items, you can use the
    `maximum_offset` parameter to limit the number of items to fetch. For example:

        client = RESTClient(
            base_url="https://api.example.com",
            paginator=OffsetPaginator(
                limit=100,
                maximum_offset=1000,
                total_path=None
            )
        )
        ...

    In this case, pagination will stop after fetching 1000 items.

    If the API provides a boolean value indicating whether there are more items
    to fetch, you can use the `has_more_path` parameter to stop the pagination
    when there are no more items to fetch. For example:

        {
            "items": [...],
            "has_more": false
        }

        client = RESTClient(
            base_url="https://api.example.com",
            paginator=OffsetPaginator(
                limit=100,
                total_path=None,
                has_more_path="has_more"
            )
        )
        ...

    In this case, pagination will stop once the `has_more` value is `false`.
    """

    def __init__(
        self,
        limit: int,
        offset: int = 0,
        offset_param: Optional[str] = None,
        limit_param: Optional[str] = None,
        total_path: Optional[jsonpath.TJsonPath] = "total",
        maximum_offset: Optional[int] = None,
        stop_after_empty_page: Optional[bool] = True,
        *,
        has_more_path: Optional[jsonpath.TJsonPath] = None,
        offset_body_path: Optional[str] = None,
        limit_body_path: Optional[str] = None,
    ) -> None:
        """
        Args:
            limit (int): The maximum number of items to retrieve
                in each request.
            offset (int): The offset for the first request.
                Defaults to 0.
            offset_param (str): The query parameter name for the offset.
                Defaults to 'offset'.
            offset_body_path (str): A dot-separated path specifying
                where to place the offset in the request JSON body.
                If provided, the paginator will use this instead of `offset_param`
                to send the offset in the request body. Defaults to `None`.
            limit_param (str): The query parameter name for the limit.
                Defaults to 'limit'.
            limit_body_path (str): A dot-separated path specifying
                where to place the limit in the request JSON body.
                If provided, the paginator will use this instead of `limit_param`
                to send the limit in the request body. Defaults to `None`.
            total_path (jsonpath.TJsonPath): The JSONPath expression for
                the total number of items.
            maximum_offset (int): The maximum offset value. If provided,
                pagination will stop once this offset is reached or exceeded,
                even if more data is available. This allows you to limit the
                maximum range for pagination. Defaults to None.
            stop_after_empty_page (bool): Whether pagination should stop when
              a page contains no result items. Defaults to `True`.
            has_more_path (jsonpath.TJsonPath): The JSONPath expression for
                the boolean value indicating whether there are more items to fetch.
                Defaults to None.
        """
        # For backward compatibility: set default offset_param if both are None
        if offset_param is None and offset_body_path is None:
            offset_param = "offset"
        # Check that only one of offset_param or offset_body_path is provided
        elif (not offset_param and not offset_body_path) or (offset_param and offset_body_path):
            raise ValueError(
                "Either 'offset_param' or 'offset_body_path' must be provided, not both."
            )

        if limit_param is None and limit_body_path is None:
            limit_param = "limit"
        # Check that only one of limit_param or limit_body_path is provided
        elif (not limit_param and not limit_body_path) or (limit_param and limit_body_path):
            raise ValueError(
                "Either 'limit_param' or 'limit_body_path' must be provided, not both."
            )

        if (
            total_path is None
            and maximum_offset is None
            and has_more_path is None
            and not stop_after_empty_page
        ):
            raise ValueError(
                "One of `total_path`, `maximum_offset`, `has_more_path`, or `stop_after_empty_page`"
                " must be provided."
            )
        super().__init__(
            param_name=offset_param,
            initial_value=offset,
            total_path=total_path,
            value_step=limit,
            maximum_value=maximum_offset,
            stop_after_empty_page=stop_after_empty_page,
            has_more_path=has_more_path,
            param_body_path=offset_body_path,
        )
        self.limit_param = limit_param
        self.limit_body_path = limit_body_path
        self.limit = limit

    def init_request(self, request: Request) -> None:
        super().init_request(request)
        self._update_request_limit(request)

    def update_request(self, request: Request) -> None:
        super().update_request(request)
        self._update_request_limit(request)

    def _update_request_limit(self, request: Request) -> None:
        if self.limit_body_path:
            self._update_request_with_body_path(request, self.limit_body_path, self.limit)
        else:
            self._update_request_with_param_name(request, self.limit_param, self.limit)

    def __str__(self) -> str:
        return (
            super().__str__()
            + f": current offset: {self.current_value} "
            + (f"offset_param: {self.param_name} " if self.param_name else "")
            + (f"offset_body_path: {self.param_body_path} " if self.param_body_path else "")
            + f"limit: {self.value_step} "
            + (f"limit_param: {self.limit_param} " if self.limit_param else "")
            + (f"limit_body_path: {self.limit_body_path} " if self.limit_body_path else "")
            + f"total_path: {self.total_path} "
            + f"maximum_value: {self.maximum_value} has_more_path: {self.has_more_path}"
        )


class BaseReferencePaginator(BasePaginator):
    """A base paginator class for paginators that use a reference to the next
    page, such as a URL or a cursor string.

    Subclasses should implement:
      1. `update_state` method to extract the next page reference and
        set the `_next_reference` attribute accordingly.
      2. `update_request` method to update the request object with the next
        page reference.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__next_reference: Optional[str] = None

    @property
    def _next_reference(self) -> Optional[str]:
        """The reference to the next page, such as a URL or a cursor.

        Returns:
            Optional[str]: The reference to the next page if available,
                otherwise None.
        """
        return self.__next_reference

    @_next_reference.setter
    def _next_reference(self, value: Optional[str]) -> None:
        """Sets the reference to the next page and updates the availability
        of the next page.

        Args:
            value (Optional[str]): The reference to the next page.
        """
        self.__next_reference = value
        self._has_next_page = value is not None


class BaseNextUrlPaginator(BaseReferencePaginator):
    """
    A base paginator class for paginators that use a URL provided in the API
    response to fetch the next page. For example, the URL can be found in HTTP
    headers or in the JSON response.

    Subclasses should implement the `update_state` method to extract the next
    page URL and set the `_next_reference` attribute accordingly.

    See `HeaderLinkPaginator` and `JSONLinkPaginator` for examples.
    """

    def update_request(self, request: Request) -> None:
        # Handle relative URLs
        if self._next_reference:
            parsed_url = urlparse(self._next_reference)
            if not parsed_url.scheme:
                self._next_reference = urljoin(request.url, self._next_reference)

        request.url = self._next_reference

        # Clear the query parameters from the previous request otherwise they
        # will be appended to the next URL in Session.prepare_request
        request.params = None


class HeaderLinkPaginator(BaseNextUrlPaginator):
    """A paginator that uses the 'Link' header in HTTP responses
    for pagination.

    A good example of this is the GitHub API:
        https://docs.github.com/en/rest/guides/traversing-with-pagination

    For example, consider an API response that includes 'Link' header:

        ...
        Content-Type: application/json
        Link: <https://api.example.com/items?page=2>; rel="next", <https://api.example.com/items?page=1>; rel="prev"

        [
            {"id": 1, "name": "item1"},
            {"id": 2, "name": "item2"},
            ...
        ]

    In this scenario, the URL for the next page (`https://api.example.com/items?page=2`)
    is identified by its relation type `rel="next"`. `HeaderLinkPaginator` extracts
    this URL from the 'Link' header and uses it to fetch the next page of results:

        from dlt.sources.helpers.rest_client import RESTClient
        client = RESTClient(
            base_url="https://api.example.com",
            paginator=HeaderLinkPaginator()
        )

        @dlt.resource
        def get_issues():
            for page in client.paginate("/items"):
                yield page
    """

    def __init__(self, links_next_key: str = "next") -> None:
        """
        Args:
            links_next_key (str, optional): The key (rel) in the 'Link' header
                that contains the next page URL. Defaults to 'next'.
        """
        super().__init__()
        self.links_next_key = links_next_key

    def update_state(self, response: Response, data: Optional[List[Any]] = None) -> None:
        """Extracts the next page URL from the 'Link' header in the response."""
        self._next_reference = response.links.get(self.links_next_key, {}).get("url")

    def __str__(self) -> str:
        return super().__str__() + f": links_next_key: {self.links_next_key}"


class JSONLinkPaginator(BaseNextUrlPaginator):
    """Locates the next page URL within the JSON response body. The key
    containing the URL can be specified using a JSON path.

    For example, suppose the JSON response from an API contains data items
    along with a 'pagination' object:

        {
            "items": [
                {"id": 1, "name": "item1"},
                {"id": 2, "name": "item2"},
                ...
            ],
            "pagination": {
                "next": "https://api.example.com/items?page=2"
            }
        }

    The link to the next page (`https://api.example.com/items?page=2`) is
    located in the 'next' key of the 'pagination' object. You can use
    `JSONLinkPaginator` to paginate through the API endpoint:

        from dlt.sources.helpers.rest_client import RESTClient
        client = RESTClient(
            base_url="https://api.example.com",
            paginator=JSONLinkPaginator(next_url_path="pagination.next")
        )

        @dlt.resource
        def get_data():
            for page in client.paginate("/posts"):
                yield page
    """

    def __init__(
        self,
        next_url_path: jsonpath.TJsonPath = "next",
    ):
        """
        Args:
            next_url_path (jsonpath.TJsonPath): The JSON path to the key
                containing the next page URL in the response body.
                Defaults to 'next'.
        """
        super().__init__()
        self.next_url_path = jsonpath.compile_path(next_url_path)

    def update_state(self, response: Response, data: Optional[List[Any]] = None) -> None:
        """Extracts the next page URL from the JSON response."""
        values = jsonpath.find_values(self.next_url_path, response.json())
        self._next_reference = values[0] if values else None

    def __str__(self) -> str:
        return super().__str__() + f": next_url_path: {self.next_url_path}"


class JSONResponsePaginator(JSONLinkPaginator):
    def __init__(
        self,
        next_url_path: jsonpath.TJsonPath = "next",
    ) -> None:
        warnings.warn(
            "JSONResponsePaginator is deprecated and will be removed in version 1.0.0. Use"
            " JSONLinkPaginator instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(next_url_path)


class JSONResponseCursorPaginator(BaseReferencePaginator):
    """Uses a cursor parameter for pagination, with the cursor value found in
    the JSON response body.

    For example, suppose the JSON response from an API contains
    a 'cursors' object:

        {
            "items": [
                {"id": 1, "name": "item1"},
                {"id": 2, "name": "item2"},
                ...
            ],
            "cursors": {
                "next": "aW1wb3J0IGFudGlncmF2aXR5"
            }
        }

    And the API endpoint expects a 'cursor' query parameter to fetch
    the next page. So the URL for the next page would look
    like `https://api.example.com/items?cursor=aW1wb3J0IGFudGlncmF2aXR5`.

    You can paginate through this API endpoint using
    `JSONResponseCursorPaginator`:

        from dlt.sources.helpers.rest_client import RESTClient
        client = RESTClient(
            base_url="https://api.example.com",
            paginator=JSONResponseCursorPaginator(
                cursor_path="cursors.next",
                cursor_param="cursor"
            )
        )

        @dlt.resource
        def get_data():
            for page in client.paginate("/posts"):
                yield page

    For requests with a JSON body, you can specify where to place the cursor in the request body
    using the `cursor_body_path` parameter. For example:

        from dlt.sources.helpers.rest_client import RESTClient
        client = RESTClient(
            base_url="https://api.example.com",
            paginator=JSONResponseCursorPaginator(
                cursor_path="nextPageToken",
                cursor_body_path="nextPageToken"
            )
        )

        @dlt.resource
        def get_data():
            for page in client.paginate(path="/search", method="POST", json={"query": "example"}):
                yield page
    """

    def __init__(
        self,
        cursor_path: jsonpath.TJsonPath = "cursors.next",
        cursor_param: Optional[str] = None,
        cursor_body_path: Optional[str] = None,
        *,
        stop_after_empty_page: bool = False,
        has_more_path: Optional[jsonpath.TJsonPath] = None,
    ):
        """
        Args:
            cursor_path: The JSON path to the key that contains the cursor in
                the response.
            cursor_param: The name of the query parameter to be used in
                the request to get the next page.
            cursor_body_path: The dot-separated path where to place the cursor in the request body.
            stop_after_empty_page: Whether pagination should stop when
                a page contains no result items. Defaults to `False`.
            has_more_path: The JSON path to a boolean value in the response
                indicating whether there are more items to fetch.
        """
        super().__init__()
        self.cursor_path = jsonpath.compile_path(cursor_path)

        # For backward compatibility: set default cursor_param if both are None
        if cursor_param is None and cursor_body_path is None:
            cursor_param = "cursor"
        # Check that only one of cursor_param or cursor_body_path is provided
        elif cursor_param is not None and cursor_body_path is not None:
            raise ValueError(
                "Either 'cursor_param' or 'cursor_body_path' must be provided, not both."
            )

        self.cursor_param = cursor_param
        self.cursor_body_path = cursor_body_path
        self.stop_after_empty_page = stop_after_empty_page
        self.has_more_path = jsonpath.compile_path(has_more_path) if has_more_path else None

    def update_state(self, response: Response, data: Optional[List[Any]] = None) -> None:
        """Extracts the cursor value from the JSON response."""
        response_json = response.json()
        values = jsonpath.find_values(self.cursor_path, response_json)
        self._next_reference = values[0] if values and values[0] else None

        if self.stop_after_empty_page and not data:
            self._has_next_page = False
            return

        has_more = None
        if self.has_more_path:
            values = jsonpath.find_values(self.has_more_path, response_json)
            has_more = values[0] if values else None
            if has_more is None:
                self._handle_missing_has_more(response_json)
            elif isinstance(has_more, str):
                try:
                    has_more = str2bool(has_more)
                except ValueError:
                    self._handle_invalid_has_more(has_more)
            elif not isinstance(has_more, bool):
                self._handle_invalid_has_more(has_more)

            self._has_next_page = has_more

    def _handle_invalid_has_more(self, has_more: Any) -> None:
        raise ValueError(
            f"'{self.has_more_path}' is not a `bool` in the response in"
            f" `{self.__class__.__name__}`. Expected a boolean, got `{has_more}`"
        )

    def _handle_missing_has_more(self, response_json: Dict[str, Any]) -> None:
        raise ValueError(
            f"Has more value not found in the response in `{self.__class__.__name__}`. "
            f"Expected a response with a `{self.has_more_path}` key, got"
            f" `{response_json}`."
        )

    def update_request(self, request: Request) -> None:
        """Updates the request with the cursor value either in query parameters
        or in the request JSON body."""
        if self.cursor_body_path:
            if request.json is None:
                request.json = {}
            jsonpath.set_value_at_path(request.json, self.cursor_body_path, self._next_reference)
        else:
            if request.params is None:
                request.params = {}
            request.params[self.cursor_param] = self._next_reference

    def __str__(self) -> str:
        return (
            super().__str__()
            + f": cursor_path: {self.cursor_path}"
            + (f" cursor_param: {self.cursor_param}" if self.cursor_param else "")
            + (f" cursor_body_path: {self.cursor_body_path}" if self.cursor_body_path else "")
        )


class HeaderCursorPaginator(BaseReferencePaginator):
    """A paginator that uses a cursor in the HTTP responses header
    for pagination.

    For example, consider an API response that includes 'NextPageToken' header:

        ...
        Content-Type: application/json
        NextPageToken: 123456"

        [
            {"id": 1, "name": "item1"},
            {"id": 2, "name": "item2"},
            ...
        ]

    In this scenario, the parameter to construct the URL for the next page (`https://api.example.com/items?page=123456`)
    is identified by the `NextPageToken` header. `HeaderCursorPaginator` extracts
    this parameter from the header and uses it to fetch the next page of results:

        from dlt.sources.helpers.rest_client import RESTClient
        client = RESTClient(
            base_url="https://api.example.com",
            paginator=HeaderCursorPaginator()
        )

        @dlt.resource
        def get_issues():
            for page in client.paginate("/items"):
                yield page
    """

    def __init__(self, cursor_key: str = "next", cursor_param: str = "cursor") -> None:
        """
        Args:
            cursor_key (str, optional): The key in the header
                that contains the next page cursor value. Defaults to 'next'.
            cursor_param (str, optional): The param name to pass the token to
                for the next request. Defaults to 'cursor'.
        """
        super().__init__()
        self.cursor_key = cursor_key
        self.cursor_param = cursor_param

    def update_state(self, response: Response, data: Optional[List[Any]] = None) -> None:
        """Extracts the next page cursor from the header in the response."""
        self._next_reference = response.headers.get(self.cursor_key)

    def update_request(self, request: Request) -> None:
        """Updates the request with the cursor query parameter."""
        if request.params is None:
            request.params = {}

        request.params[self.cursor_param] = self._next_reference

    def __str__(self) -> str:
        return (
            super().__str__()
            + f": cursor_value: {self._next_reference} cursor_param: {self.cursor_param}"
        )



# TODO: register layout types and specs for each top level sections so we can generate full config typing

SOURCES = "sources"
"""a top section holding source and resource configs often within their own sections named after modules they are in"""

DESTINATION = "destination"
"""a top section holding sections named after particular destinations with configurations and credentials. NOTE: will be deprecated"""

DESTINATIONS = "destinations"
"""a top section holding sections named after particular destinations with configurations and credentials. NOTE: not yet supported"""

PIPELINES = "pipelines"
"""a top section holding pipeline configurations"""

DATASETS = "datasets"
"""a top section holding dataset configurations"""

LOAD = "load"
"""load and load storage configuration"""

NORMALIZE = "normalize"
"""normalize and normalize storage configuration"""

EXTRACT = "extract"
"""extract stage of the pipeline"""

SCHEMA = "schema"
"""schema configuration, ie. normalizers"""

PROVIDERS = "providers"
"""secrets and config providers"""

DATA_WRITER = "data_writer"
"""default section holding BufferedDataWriter settings"""

DBT_PACKAGE_RUNNER = "dbt_package_runner"
"""dbt package runner configuration (DBTRunnerConfiguration)"""

DBT_CLOUD = "dbt_cloud"
"""dbt cloud helpers configuration (DBTCloudConfiguration)"""

DASHBOARD = "dashboard"
"""dashboard configuration (DashboardConfiguration)"""

WORKSPACE = "workspace"
"""Workspace configuration"""



from typing import ClassVar

from dlt.common.normalizers.naming.naming import NamingConvention as BaseNamingConvention


class NamingConvention(BaseNamingConvention):
    """Case sensitive naming convention that maps source identifiers to destination identifiers with
    only minimal changes. New line characters, double and single quotes are replaced with underscores.

    Uses ▶ as path separator.
    """

    PATH_SEPARATOR: ClassVar[str] = "▶"
    _CLEANUP_TABLE = str.maketrans("\n\r'\"▶", "_____")

    def normalize_identifier(self, identifier: str) -> str:
        identifier = super().normalize_identifier(identifier)
        norm_identifier = identifier.translate(self._CLEANUP_TABLE)
        return self.shorten_identifier(norm_identifier, identifier, self.max_length)

    @property
    def is_case_sensitive(self) -> bool:
        return True



from dlt.common.data_writers.writers import (
    DataWriter,
    TDataItemFormat,
    FileWriterSpec,
    create_import_spec,
    resolve_best_writer_spec,
    get_best_writer_spec,
    is_native_writer,
)
from dlt.common.data_writers.buffered import BufferedDataWriter, new_file_id
from dlt.common.data_writers.escape import (
    escape_redshift_literal,
    escape_redshift_identifier,
    escape_hive_identifier,
)

__all__ = [
    "DataWriter",
    "FileWriterSpec",
    "create_import_spec",
    "resolve_best_writer_spec",
    "get_best_writer_spec",
    "is_native_writer",
    "TDataItemFormat",
    "BufferedDataWriter",
    "new_file_id",
    "escape_redshift_literal",
    "escape_redshift_identifier",
    "escape_hive_identifier",
]



import re

from typing import Any, List, Dict, Type, Optional, Sequence, Tuple, cast

from dlt.common import logger
from dlt.common.destination.capabilities import DestinationCapabilitiesContext
from dlt.common.destination.exceptions import WriteDispositionNotSupported
from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.destination.utils import resolve_merge_strategy, resolve_replace_strategy
from dlt.common.schema import Schema
from dlt.common.schema.exceptions import SchemaCorruptedException
from dlt.common.schema.typing import (
    MERGE_STRATEGIES,
    TColumnType,
    TLoaderReplaceStrategy,
    TTableSchema,
)
from dlt.common.schema.utils import (
    get_columns_names_with_prop,
    get_first_column_name_with_prop,
    has_column_with_prop,
    is_nested_table,
    pipeline_state_table,
)

from dlt.destinations.exceptions import DatabaseTransientException
from dlt.extract import DltResource, resource as make_resource, DltSource

RE_DATA_TYPE = re.compile(r"([A-Z]+)\((\d+)(?:,\s?(\d+))?\)")


def get_resource_for_adapter(data: Any) -> DltResource:
    """
    Helper function for adapters. Wraps `data` in a DltResource if it's not a DltResource already.
    Alternatively if `data` is a DltSource, throws an error if there are multiple resource in the source
    or returns the single resource if available.
    """
    if isinstance(data, DltResource):
        return data
    # prevent accidentally wrapping sources with adapters
    if isinstance(data, DltSource):
        if len(data.selected_resources.keys()) == 1:
            return list(data.selected_resources.values())[0]
        else:
            raise ValueError(
                "You are trying to use an adapter on a `DltSource` with multiple resources. You can"
                " only use adapters on: pure data, a `DltResouce` or a `DltSource` with a single"
                " `DltResource`."
            )

    resource_name = None
    if not hasattr(data, "__name__"):
        logger.info("Setting default resource name to `content` for adapted resource.")
        resource_name = "content"

    return cast(DltResource, make_resource(data, name=resource_name))


def info_schema_null_to_bool(v: str) -> bool:
    """Converts INFORMATION SCHEMA truth values to Python bool"""
    if v in ("NO", "0"):
        return False
    elif v in ("YES", "1"):
        return True
    raise ValueError(v)


def parse_db_data_type_str_with_precision(db_type: str) -> Tuple[str, Optional[int], Optional[int]]:
    """Parses a db data type with optional precision or precision and scale information"""
    # Search for matches using the regular expression
    match = RE_DATA_TYPE.match(db_type)

    # If the pattern matches, extract the type, precision, and scale
    if match:
        db_type = match.group(1)
        precision = int(match.group(2))
        scale = int(match.group(3)) if match.group(3) else None
        return db_type, precision, scale

    # If the pattern does not match, return the original type without precision and scale
    return db_type, None, None


def get_pipeline_state_query_columns() -> TTableSchema:
    """We get definition of pipeline state table without columns we do not need for the query"""
    state_table = pipeline_state_table()
    # we do not need version_hash to be backward compatible as long as we can
    state_table["columns"].pop("version_hash")
    return state_table


def verify_schema_replace_disposition(
    schema: Schema,
    load_tables: Sequence[PreparedTableSchema],
    capabilities: DestinationCapabilitiesContext,
    required_strategy: TLoaderReplaceStrategy,
    warnings: bool = True,
) -> List[Exception]:
    # log = logger.warning if warnings else logger.info
    # collect all exceptions to show all problems in the schema
    exception_log: List[Exception] = []

    # verifies schema settings specific to sql job client
    for table in load_tables:
        # from now on validate only top level tables
        if is_nested_table(table):
            continue
        table_name = table["name"]
        if table["write_disposition"] == "replace":
            if not resolve_replace_strategy(table, required_strategy, capabilities):
                format_info = ""
                if capabilities.supported_table_formats:
                    format_info = (
                        " Note that setting table format may allow you to pick other replace"
                        " strategies. Available formats:"
                        f" {', '.join(capabilities.supported_table_formats)}"
                    )
                exception_log.append(
                    # Pick one of: {', '.join(supported_replace_strategies)} .
                    SchemaCorruptedException(
                        schema.name,
                        f"Requested replace strategy {required_strategy} not available for table"
                        f" {table_name}. "
                        + format_info,
                    )
                )
    return exception_log


def verify_schema_merge_disposition(
    schema: Schema,
    load_tables: Sequence[PreparedTableSchema],
    capabilities: DestinationCapabilitiesContext,
    warnings: bool = True,
) -> List[Exception]:
    log = logger.warning if warnings else logger.info
    # collect all exceptions to show all problems in the schema
    exception_log: List[Exception] = []

    # verifies schema settings specific to sql job client
    for table in load_tables:
        # from now on validate only top level tables
        if is_nested_table(table):
            continue

        table_name = table["name"]
        if table["write_disposition"] == "merge":
            if "x-merge-strategy" in table and table["x-merge-strategy"] not in MERGE_STRATEGIES:  # type: ignore[typeddict-item]
                exception_log.append(
                    SchemaCorruptedException(
                        schema.name,
                        f'"{table["x-merge-strategy"]}" is not a valid merge strategy. '  # type: ignore[typeddict-item]
                        f"""Allowed values: {', '.join(['"' + s + '"' for s in MERGE_STRATEGIES])}.""",
                    )
                )

            merge_strategy = resolve_merge_strategy(schema.tables, table, capabilities)
            if merge_strategy is None:
                table_format_info = ""
                if capabilities.supported_table_formats:
                    table_format_info = (
                        " or try different table format which may offer `merge`:"
                        f" {capabilities.supported_table_formats}"
                    )
                exception_log.append(
                    WriteDispositionNotSupported(
                        "merge",
                        "Destination does not support any merge strategies and `merge` write"
                        f" disposition  for table `{table_name}` cannot be met. "
                        f"Change write disposition{table_format_info}.",
                    )
                )
                continue
            if merge_strategy == "delete-insert":
                if not has_column_with_prop(table, "primary_key") and not has_column_with_prop(
                    table, "merge_key"
                ):
                    log(
                        f"Table {table_name} has `write_disposition` set to `merge`"
                        " and `merge_strategy` set to `delete-insert`, but no primary or"
                        " merge keys defined."
                    )
            elif merge_strategy == "upsert":
                if not has_column_with_prop(table, "primary_key"):
                    exception_log.append(
                        SchemaCorruptedException(
                            schema.name,
                            f"No primary key defined for table `{table['name']}`."
                            " `primary_key` needs to be set when using the `upsert`"
                            " merge strategy.",
                        )
                    )
                if has_column_with_prop(table, "merge_key"):
                    log(
                        f"Found `merge_key` for table `{table['name']}` with"
                        " `upsert` merge strategy. Merge key is not supported"
                        " for this strategy and will be ignored."
                    )
        if has_column_with_prop(table, "hard_delete"):
            if len(get_columns_names_with_prop(table, "hard_delete")) > 1:
                exception_log.append(
                    SchemaCorruptedException(
                        schema.name,
                        f'Found multiple "hard_delete" column hints for table "{table_name}" in'
                        f' schema "{schema.name}" while only one is allowed:'
                        f' {", ".join(get_columns_names_with_prop(table, "hard_delete"))}.',
                    )
                )
            if table.get("write_disposition") in ("replace", "append"):
                log(
                    f"""The "hard_delete" column hint for column "{get_first_column_name_with_prop(table, 'hard_delete')}" """
                    f'in table "{table_name}" with write disposition'
                    f' "{table.get("write_disposition")}"'
                    f' in schema "{schema.name}" will be ignored.'
                    ' The "hard_delete" column hint is only applied when using'
                    ' the "merge" write disposition.'
                )
        if has_column_with_prop(table, "dedup_sort"):
            if len(get_columns_names_with_prop(table, "dedup_sort")) > 1:
                exception_log.append(
                    SchemaCorruptedException(
                        schema.name,
                        f'Found multiple "dedup_sort" column hints for table "{table_name}" in'
                        f' schema "{schema.name}" while only one is allowed:'
                        f' {", ".join(get_columns_names_with_prop(table, "dedup_sort"))}.',
                    )
                )
            if table.get("write_disposition") in ("replace", "append"):
                log(
                    f"""The "dedup_sort" column hint for column "{get_first_column_name_with_prop(table, 'dedup_sort')}" """
                    f'in table "{table_name}" with write disposition'
                    f' "{table.get("write_disposition")}"'
                    f' in schema "{schema.name}" will be ignored.'
                    ' The "dedup_sort" column hint is only applied when using'
                    ' the "merge" write disposition.'
                )
            if table.get("write_disposition") == "merge":
                if not has_column_with_prop(table, "primary_key"):
                    log(
                        f"""The "dedup_sort" column hint for column "{get_first_column_name_with_prop(table, 'dedup_sort')}" """
                        f'in table "{table_name}" with write disposition'
                        f' "{table.get("write_disposition")}"'
                        f' in schema "{schema.name}" will be ignored.'
                        ' The "dedup_sort" column hint is only applied when a'
                        " primary key has been specified."
                    )
                if table.get("x-stage-data-deduplicated", False):
                    log(
                        f"""The "dedup_sort" column hint for column "{get_first_column_name_with_prop(table, 'dedup_sort')}" """
                        f'in table "{table_name}" with write disposition'
                        f' "{table.get("write_disposition")}"'
                        f' in schema "{schema.name}" will be ignored.'
                        " Staging data is declared as already deduplicated."
                    )
    return exception_log


def _convert_to_old_pyformat(
    new_style_string: str, args: Tuple[Any, ...], operational_error_cls: Type[Exception]
) -> Tuple[str, Dict[str, Any]]:
    """Converts a query string from the new pyformat style to the old pyformat style.

    The new pyformat style uses placeholders like %s, while the old pyformat style
    uses placeholders like %(arg0)s, where the number corresponds to the index of
    the argument in the args tuple.

    Args:
        new_style_string (str): The query string in the new pyformat style.
        args (Tuple[Any, ...]): The arguments to be inserted into the query string.
        operational_error_cls (Type[Exception]): The specific OperationalError class to be raised
            in case of a mismatch between placeholders and arguments. This should be the
            OperationalError class provided by the DBAPI2-compliant driver being used.

    Returns:
        Tuple[str, Dict[str, Any]]: A tuple containing the converted query string
            in the old pyformat style, and a dictionary mapping argument keys to values.

    Raises:
        DatabaseTransientException: If there is a mismatch between the number of
            placeholders in the query string, and the number of arguments provided.
    """
    if len(args) == 1 and isinstance(args[0], tuple):
        args = args[0]

    keys = [f"arg{str(i)}" for i, _ in enumerate(args)]
    old_style_string, count = re.subn(r"%s", lambda _: f"%({keys.pop(0)})s", new_style_string)
    mapping = dict(zip([f"arg{str(i)}" for i, _ in enumerate(args)], args))
    if count != len(args):
        raise DatabaseTransientException(operational_error_cls())
    return old_style_string, mapping


def is_compression_disabled() -> bool:
    from dlt import config

    key_ = "normalize.data_writer.disable_compression"
    if key_ not in config:
        key_ = "data_writer.disable_compression"
    return config.get(key_, bool)


def get_deterministic_temp_table_name(table_name: str, op: str) -> str:
    """Returns table name suitable for deterministic temp table. Such table will survive
    disconnects and is supposed to be dropped before being filled again so we need
    deterministic name with very low collision prob with "data" tables
    """
    from dlt.common.normalizers.naming import NamingConvention

    op_name = f"{table_name}_{op}"
    return f"{op_name}_{NamingConvention._compute_tag(op_name, 0.001)}"



"""Dashboard UI helpers: tables, callouts, section boilerplate, and shared utilities."""

from typing import Any, Dict, List, Optional, Tuple, Union


class BoundedDict(Dict[Any, Any]):
    """Dict that evicts the oldest entry when *maxsize* is exceeded."""

    def __init__(self, maxsize: int = 64) -> None:
        super().__init__()
        self._maxsize = maxsize

    def __setitem__(self, key: Any, value: Any) -> None:
        if key not in self and len(self) >= self._maxsize:
            del self[next(iter(self))]
        super().__setitem__(key, value)


import marimo as mo
import pyarrow
import traceback


import dlt

from dlt._workspace.helpers.dashboard import strings
from dlt._workspace.helpers.dashboard.strings import TSectionStrings
from dlt._workspace.helpers.dashboard.utils.formatters import align_dict_keys


def _style_cell(row_id: str, name: str, __: Any) -> Dict[str, str]:
    """Apply alternating row colors and bold the name column."""
    style: Dict[str, str] = {"background-color": "white" if (int(row_id) % 2 == 0) else "#f4f4f9"}
    if name.lower() == "name":
        style["font-weight"] = "bold"
    return style


def dlt_table(
    data: Union[List[Dict[str, Any]], pyarrow.Table],
    *,
    selection: Optional[str] = None,
    style: bool = True,
    freeze_column: Optional[str] = "name",
    initial_selection: Optional[List[int]] = None,
    pagination: Optional[bool] = None,
    show_download: Optional[bool] = None,
    **kwargs: Any,
) -> mo.ui.table:
    """Create a styled mo.ui.table with common dashboard defaults.

    Applies alternating-row styling, freezes the given column, aligns dict keys,
    and sets initial_selection to None when data is empty.

    Args:
        data: Table data as a list of dicts or a pyarrow Table.
        selection: Selection mode ("single", "multi") or None to disable.
        style: Whether to apply alternating-row cell styling.
        freeze_column: Column name to freeze on the left. Pass None to disable.
        initial_selection: Row indices to pre-select. Forced to None when data is empty.
        pagination: Passed through to mo.ui.table.
        show_download: Passed through to mo.ui.table.
    """
    is_list = isinstance(data, list)
    has_data = (len(data) > 0) if is_list else (data.num_rows > 0)  # type: ignore[union-attr]

    # align dict keys for consistent column widths
    if is_list and has_data:
        data = align_dict_keys(data)

    # build kwargs for mo.ui.table
    table_kwargs: Dict[str, Any] = {"selection": selection, **kwargs}

    if style:
        table_kwargs["style_cell"] = _style_cell

    if freeze_column is not None and has_data and freeze_column in data[0]:
        table_kwargs["freeze_columns_left"] = [freeze_column]

    if initial_selection is not None and has_data:
        table_kwargs["initial_selection"] = initial_selection

    if pagination is not None:
        table_kwargs["pagination"] = pagination

    if show_download is not None:
        table_kwargs["show_download"] = show_download

    return mo.ui.table(data, **table_kwargs)


def small(text: str) -> str:
    """Wrap text in <small> HTML tags for consistent dashboard styling."""
    return f"<small>{text}</small>"


def error_callout(message: str, code: str = None, traceback_string: str = None) -> mo.Html:
    """Build a callout with a message and a exposable stacktrace.

    Args:
        message (str): The message to display in the callout.
        code (str): The code to display in the callout.
        traceback_string (str): The traceback to display in the callout.

    Returns:
        mo.ui.Callout: The callout with the message and the stacktrace.
    """
    if code:
        code = code.replace("\x1b[4m", ">>>").replace("\x1b[0m", "<<<")

    traceback_string = traceback_string or traceback.format_exc()
    if traceback_string.startswith("NoneType: None"):
        traceback_string = None

    stack_items = [mo.md(message)]
    if code:
        stack_items.append(
            mo.ui.code_editor(code, language="python", disabled=True, show_copy_button=True)
        )
    if traceback_string:
        stack_items.append(
            mo.accordion(
                {
                    strings.error_show_stacktrace: mo.ui.code_editor(
                        traceback_string, language="python", disabled=True, show_copy_button=True
                    )
                }
            )
        )

    return mo.callout(
        mo.vstack(stack_items),
        kind="warn",
    )


def title_and_subtitle(title: str, subtitle: str = None, title_level: int = 3) -> mo.Html:
    """Build a title and a subtitle block"""
    _result = []
    if title:
        _result.append(mo.md(f"{title_level * '#'} {title}"))
    if subtitle:
        _result.append(mo.md(small(subtitle)))
    return mo.vstack(_result)


def page_header(
    dlt_pipeline: dlt.Pipeline,
    section_strings: TSectionStrings,
    button: mo.ui.switch = None,
) -> List[mo.Html]:
    """Build a page header with a title, a subtitle, button and conditional longer subtitle.

    When collapsed, the short subtitle is shown inline after the title to save
    vertical space.  When expanded, the longer subtitle appears on its own line.
    """
    if not dlt_pipeline:
        return []
    if button.value:
        title_block = title_and_subtitle(
            section_strings.title, section_strings.subtitle_long, title_level=2
        )
        header = mo.hstack([title_block, button], align="center")
        css_class = "section-header"
    else:
        # title left, subtitle center, switch right
        subtitle_html = mo.Html(
            f'<span class="section-subtitle">{small(section_strings.subtitle)}</span>'
        )
        switch_html = mo.Html(f'<span class="section-switch">{button.text}</span>')
        header = mo.hstack(
            [mo.md(f"## {section_strings.title}"), subtitle_html, switch_html],
            align="center",
        )
        css_class = "section-header"
    return [mo.Html(f'<div class="{css_class}">{header.text}</div>')]


def section_marker(section_name: str, has_content: bool = False) -> mo.Html:
    """Create an invisible marker element to identify sections for CSS styling.

    Args:
        section_name: Name identifier for the section (e.g., "home_section", "schema_section")
        has_content: If True, adds 'has-content' class to enable CSS styling (borders, backgrounds).
                     Should be True only when the section has actual content and is displayed.

    Returns:
        Hidden HTML div element with section marker classes for CSS targeting.
    """
    content_class = "has-content" if has_content else ""
    return mo.Html(
        f'<div class="section-marker {content_class}" data-section="{section_name}"'
        ' hidden="hidden"></div>'
    )


def section(
    section_strings: TSectionStrings,
    dlt_pipeline: dlt.Pipeline,
    switch: mo.ui.switch,
) -> Tuple[List[mo.Html], bool]:
    """Build standard section boilerplate: marker + page header.

    Returns:
        Tuple of (result list, should_render_content). The list already
        contains the section marker and header.  Append content only when
        should_render_content is True, then call `mo.vstack(result)`.
    """
    result = [
        section_marker(section_strings.section_name, has_content=dlt_pipeline is not None),
    ]
    result.extend(page_header(dlt_pipeline, section_strings, switch))
    return result, bool(dlt_pipeline and switch.value)



import datetime  # noqa: I251
from typing import Any, Dict, List, NamedTuple, Optional, Tuple  # noqa: 251
from dlt.common.typing import TypedDict


class DataWriterMetrics(NamedTuple):
    file_path: str
    items_count: int
    file_size: int
    created: float
    last_modified: float

    def __add__(self, other: Tuple[object, ...], /) -> Tuple[object, ...]:  # type: ignore[override]
        if isinstance(other, DataWriterMetrics):
            return DataWriterMetrics(
                self.file_path if self.file_path == other.file_path else "",
                # self.table_name if self.table_name == other.table_name else "",
                self.items_count + other.items_count,
                self.file_size + other.file_size,
                min(self.created, other.created),
                max(self.last_modified, other.last_modified),
            )
        return NotImplemented


class DataWriterAndCustomMetrics(DataWriterMetrics):
    custom_metrics: Dict[str, Any]

    def __new__(
        cls,
        file_path: str,
        items_count: int,
        file_size: int,
        created: float,
        last_modified: float,
        custom_metrics: Dict[str, Any] = None,
    ) -> "DataWriterAndCustomMetrics":
        self = super(DataWriterAndCustomMetrics, cls).__new__(
            cls, file_path, items_count, file_size, created, last_modified
        )
        self.custom_metrics = custom_metrics or {}
        return self

    def _asdict(self) -> Dict[str, Any]:
        """Override _asdict to include custom_metrics in serialization"""
        result = super()._asdict()
        result["custom_metrics"] = self.custom_metrics
        return result


class StepMetrics(TypedDict):
    """Metrics for particular package processed in particular pipeline step"""

    started_at: datetime.datetime
    """Start of package processing"""
    finished_at: datetime.datetime
    """End of package processing"""


class ExtractDataInfo(TypedDict):
    name: str
    data_type: str


class ExtractMetrics(StepMetrics):
    schema_name: str
    job_metrics: Dict[str, DataWriterMetrics]
    """Metrics collected per job id during writing of job file"""
    table_metrics: Dict[str, DataWriterMetrics]
    """Job metrics aggregated by table"""
    resource_metrics: Dict[str, DataWriterAndCustomMetrics]
    """Job metrics aggregated by resource"""
    dag: List[Tuple[str, str]]
    """A resource dag where elements of the list are graph edges"""
    hints: Dict[str, Dict[str, Any]]
    """Hints passed to the resources"""


class NormalizeMetrics(StepMetrics):
    job_metrics: Dict[str, DataWriterMetrics]
    """Metrics collected per job id during writing of job file"""
    table_metrics: Dict[str, DataWriterMetrics]
    """Job metrics aggregated by table"""


class LoadJobMetrics(NamedTuple):
    job_id: str
    file_path: str
    table_name: str
    started_at: datetime.datetime
    finished_at: Optional[datetime.datetime]
    state: str
    remote_url: Optional[str]
    retry_count: int = 0


class LoadMetrics(StepMetrics):
    job_metrics: Dict[str, LoadJobMetrics]
